---------------------------------------------------------------------------
--                                                                       --
--  Ast2Cfg is an ASIS application that generates a CFG out of the AST.  --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--     Copyright (c) 2007, 2008, 2009 Raul Fechete, Georg Kienesberger   --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of Ast2Cfg.                                         --
--                                                                       --
-- Ast2Cfg is free software; you can redistribute it and/or modify       --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- Ast2Cfg is distributed in the hope that it will be useful,            --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ada.Unchecked_Deallocation;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Ada.Finalization; use Ada.Finalization;

with Asis;

with Ast2Cfg.Id;
with Ast2Cfg.Declaration_Id;
with Ast2Cfg.Generic_Pointer_List;
with Ast2Cfg.Utils; use Ast2Cfg.Utils;

package Ast2Cfg.Flow_Types is

   type Flow_Object    is abstract tagged limited private;
   type Flow_Class_Ptr is access all Flow_Object'Class;

   procedure Free (Flow: in out Flow_Class_Ptr);
   package Flow_Class_Ptr_List is new Generic_Pointer_List(Item_Pointer => Flow_Class_Ptr,
                                                           Free_Procedure => Free);

   type Declaration;
   type Declaration_Ptr is access all Declaration;
   procedure Free (Decl: in out Declaration_Ptr);
   package Declaration_Ptr_List is new Generic_Pointer_List(Item_Pointer => Declaration_Ptr,
                                                            Free_Procedure => Free);

   -- Type Description

   type Type_Descr is
      record
         Type_Name:      Unbounded_String;   -- The Name of the Type
         Super_Type:     UStr_List.Object;   -- Name of the Super Type (Inheritance)
         Descr_Tree:     Flow_Class_Ptr;     -- Parameter Tree for Expressions that may be used
         Component_Type: UStr_List.Object;   -- Base Type (in Aggregate Components)
         Used_Idents:    UStr_List_List.Object;       -- Identifiers found on the Way
         Discriminants:  Declaration_Ptr_List.Object; -- Type Discriminants
         Components:     Declaration_Ptr_List.Object; -- Type Components
         Is_Private:     Boolean;                     -- Is this Type defined as "Private"?
      end record;
   type Type_Descr_Ptr is access all Type_Descr;
   procedure Clear (TDescr: in out Type_Descr);
   procedure Free (TDescr: in out Type_Descr_Ptr);
   package Type_Descr_Ptr_List is new Generic_Pointer_List(Item_Pointer => Type_Descr_Ptr,
                                                          Free_Procedure => Free);

   -- Declarations

   type Declaration is new Controlled with
      record
         Decl_Id: Declaration_Id.Object; -- unique id for tracking the order of declarations
         Vars:    UStr_List.Object;      -- String List of the declared Variables
         VType:   aliased Type_Descr;    -- Type Description Object for the Variable Type
         Init:    Flow_Class_Ptr;        -- Pointer to the Root Node of a Param Tree (Initialization)
      end record;
   type Declaration_Class_Ptr is access all Declaration'Class;

   overriding
   procedure Initialize (Decl: in out Declaration);

   -- Renamings / Aliases

   type Alias is abstract new Controlled with
      record
         Decl_Id: Declaration_Id.Object; -- unique id for tracking the order of declarations
         New_Name: Unbounded_String;
         Old_Name: UStr_List.Object;
         PTree:    Flow_Class_Ptr;
      end record;

   overriding
   procedure Initialize (A: in out Alias);

   type Proc_Alias    is new Alias with null record;
   type Func_Alias    is new Alias with null record;
   type Pkg_Alias     is new Alias with null record;
   type Obj_Alias     is new Alias with null record;
   type Except_Alias  is new Alias with null record;
   type Generic_Alias is new Alias with null record;
   type Alias_Class_Ptr is access all Alias'Class;

   procedure Free (Alias: in out Alias_Class_Ptr);
   package Alias_Class_Ptr_List is new Generic_Pointer_List (Item_Pointer => Alias_Class_Ptr,
                                                             Free_Procedure => Free);

   type Instance_Alias is abstract new Alias with
      record
         Assocs: Alias_Class_Ptr_List.Object;
      end record;
   type Instance_Alias_Class_Ptr is access all Instance_Alias'Class;
   type Pkg_Inst_Alias  is new Instance_Alias with null record;
   type Proc_Inst_Alias is new Instance_Alias with null record;
   type Func_Inst_Alias is new Instance_Alias with null record;

   -- Formal Declarations

   type Formal_Decl is abstract new Declaration with
      record
         Assocs: Alias_Class_Ptr_List.Object; -- A formal Declaration may have generic Associations
      end record;
   type Formal_Decl_Class_Ptr is access all Formal_Decl'Class;
   procedure Free (FDecl: in out Formal_Decl_Class_Ptr);
   package Formal_Decl_Class_Ptr_List is new Generic_Pointer_List (Item_Pointer => Formal_Decl_Class_Ptr,
                                                                   Free_Procedure => Free);
   type Pkg_Formal_Decl  is new Formal_Decl with null record;
   type Obj_Formal_Decl  is new Formal_Decl with null record;
   type Proc_Formal_Decl is new Formal_Decl with null record;
   type Func_Formal_Decl is new Formal_Decl with null record;
   type Type_Formal_Decl is new Formal_Decl with null record;

   procedure Init      (Flow: in out Flow_Object;
                        Oid:  in Id.Pointer;
                        Name: in Unbounded_String;
                        Pred: in Flow_Class_Ptr := null;
                        Element: in Asis.Element := Asis.Nil_Element
                       );

   function  Get_Id    (Flow: in Flow_Object) return Id.Pointer;

   procedure Set_Name  (Flow: in out Flow_Object; Name: in Unbounded_String);
   function  Get_Name  (Flow: in Flow_Object) return String;
   function  Get_Name  (Flow: in Flow_Object) return Unbounded_String;

   procedure Set_Prefix (Flow: in out Flow_Object; Prefix: in out UStr_List.Object);
   function  Get_Prefix (Flow: in Flow_Object) return Unbounded_String;
   procedure Get_Prefix (Flow: in Flow_Object; Prefix: out UStr_List.Object);

   -- get the full name of the flow object or its prefix in case it has no name
   function  Get_Full_Name (Flow: in Flow_Object) return Unbounded_String;
   procedure Get_Full_Name (Flow: in Flow_Object; Full_Name: out UStr_List.Object);

   function Get_Element (Flow: in Flow_Object) return Asis.Element;

   procedure Set_PTree (Flow: in out Flow_Object; PTree: in Flow_Class_Ptr);
   function  Get_PTree (Flow: in Flow_Object) return Flow_Class_Ptr;

   procedure Add_Pred  (Flow: in out Flow_Object; Pred: in Flow_Class_Ptr);
   procedure Rem_Pred  (Flow: in out Flow_Object; Pred: in Flow_Class_Ptr);
   function  Get_Preds (Flow: in Flow_Object) return Flow_Class_Ptr_List.Object;

   procedure Add_Succ  (Flow: in out Flow_Object; Succ: in Flow_Class_Ptr);
   procedure Rem_Succ  (Flow: in out Flow_Object; Succ: in Flow_Class_Ptr);
   function  Get_Succs (Flow: in Flow_Object) return Flow_Class_Ptr_List.Object;
   procedure Clear_Succs (Flow: in out Flow_Object);

   procedure Rem_Last_Succ (Flow: in out Flow_Object);
   procedure Rem_Last_Pred (Flow: in out Flow_Object);
   function  Get_Last_Pred (Flow: in Flow_Object) return Flow_Class_Ptr;

   procedure Rem_All_Succs (Flow: in out Flow_Object);

   procedure Add_Decl  (Flow: in out Flow_Object; Decl: in Declaration_Ptr);
   function  Get_Decls (Flow: in Flow_Object) return Declaration_Ptr_List.Object;

   procedure Add_FDecl  (Flow: in out Flow_Object; FDecl: in Formal_Decl_Class_Ptr);
   procedure Set_FDecls (Flow: in out Flow_Object; FDecls: in out Formal_Decl_Class_Ptr_List.Object);

   function  Get_FDecls (Flow: in Flow_Object) return Formal_Decl_Class_Ptr_List.Object;

   procedure Add_Ren  (Flow: in out Flow_Object; Ren: in Alias_Class_Ptr);
   function  Get_Rens (Flow: in Flow_Object) return Alias_Class_Ptr_List.Object;

   procedure Add_Type  (Flow: in out Flow_Object; TDescr: in Type_Descr_Ptr);
   function  Get_Types (Flow: in Flow_Object) return Type_Descr_Ptr_List.Object;

   type Reference is abstract tagged
      record
         Name: UStr_List.Object;
      end record;
   type Ref_Class_Ptr is access all Reference'Class;
   type With_Ref is new Reference with null record;
   type Use_Pkg_Ref is new Reference with null record;
   type Use_Type_Ref is new Reference with null record;

   procedure Free (Ref: in out Ref_Class_Ptr);
   package Ref_Class_Ptr_List is new Generic_Pointer_List (Item_Pointer => Ref_Class_Ptr,
                                                           Free_Procedure => Free);

   not overriding
     procedure Add_Ref  (Flow: in out Flow_Object; Ref: in Ref_Class_Ptr);
   not overriding
     procedure Set_Refs (Flow: in out Flow_Object; Refs: in Ref_Class_Ptr_List.Object);
   not overriding
     function  Get_Refs (Flow: in Flow_Object) return Ref_Class_Ptr_List.Object;

private

   type Flow_Object is tagged limited
      record
         Oid:     Id.Pointer;
         Name:    Unbounded_String;
         PTree:   Flow_Class_Ptr;
         -- these are "Pointer"s because of GNAT's esoteric
         -- behaviour concerning deallocation
         Prefix:  UStr_List.Pointer;
         Preds:   Flow_Class_Ptr_List.Pointer;
         Succs:   Flow_Class_Ptr_List.Pointer;
         Decls:   Declaration_Ptr_List.Pointer;
         FDecls:  Formal_Decl_Class_Ptr_List.Pointer;
         Rens:    Alias_Class_Ptr_List.Pointer;
         Refs:    Ref_Class_Ptr_List.Pointer;
         Types:   Type_Descr_Ptr_List.Pointer;
         Element: Asis.Element := Asis.Nil_Element; -- The ASIS Element this flow object was derived from
      end record;

end Ast2Cfg.Flow_Types;
