---------------------------------------------------------------------------
--                                                                       --
--  Ast2Cfg is an ASIS application that generates a CFG out of the AST.  --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--     Copyright (c) 2007, 2008, 2009 Raul Fechete, Georg Kienesberger   --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of Ast2Cfg.                                         --
--                                                                       --
-- Ast2Cfg is free software; you can redistribute it and/or modify       --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- Ast2Cfg is distributed in the hope that it will be useful,            --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

-- A D A
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Ada.Characters.Handling;
with Ada.Characters.Latin_1;

-- A S I S
with Asis;
use type Asis.Element;
with Asis.Elements;
with Asis.Statements;
with Asis.Expressions;
with Asis.Text; use Asis.Text;

-- A S T 2 C F G
with Ast2Cfg.Flow_Types; use Ast2Cfg.Flow_Types;
with Ast2Cfg.Id;
with Ast2Cfg.Pkgs;
use type Ast2Cfg.Pkgs.Pkg_Class_Ptr;
with Ast2Cfg.CFGs;
use type Ast2Cfg.CFGs.CFG_Class_Ptr;
with Ast2Cfg.Nodes;
use type Ast2Cfg.Nodes.Node_Class_Ptr;
use type Ast2Cfg.Nodes.Node_Mark;
use Ast2Cfg.Nodes.Node_Class_Ptr_List;
use Ast2Cfg.Flow_Types.Alias_Class_Ptr_List;
with Ast2Cfg.Post_Trans_Processing;
with Ast2Cfg.Utils; use Ast2Cfg.Utils;
with Ast2Cfg.Output; use Ast2Cfg.Output;
with Ast2Cfg.Impl_Msg;

package body Ast2Cfg.Transformation is

   -- Initialize
   procedure Init is
   begin
      Flow_World.Init(World);
      Element_Stack.Init(EStack);
      Flow_Stack.Init(FStack);
      Param_Stack.Init(PStack);
   end Init;

   -- Finalize
   procedure Final (Interprocedural: in Boolean := False; Compress_Param_Trees: in Boolean := True) is
   begin
      Ast2Cfg.Transformation.Interprocedural := Interprocedural;
      Ast2Cfg.Transformation.Compress_Param_Trees := Compress_Param_Trees;
      Post_Trans_Process(World);
      Element_Stack.Free(EStack);
      Flow_Stack.Clear(FStack);
      Param_Stack.Free(PStack);
   end Final;

   -- Finalize and return the resulting World
   function Final (Interprocedural: in Boolean := False; Compress_Param_Trees: in Boolean := True) return Flow_World.World_Object_Ptr is
   begin
      Final(Interprocedural, Compress_Param_Trees);
      return World'Access;
   end Final;

   -- Post Transformation Processing
   procedure Post_Trans_Process (World: in out Flow_World.World_Object) is
   begin
      Put_Line(Debug, "[Transformation] Post Transformation Processing - Begin.");
      if Ast2Cfg.Transformation.Compress_Param_Trees then
         Ast2Cfg.Post_Trans_Processing.Compress_Param_Trees(World);
      else
         Put_Line(Debug, "[Transformation] Compress parameter trees option not set: Skipped compressing the parameter trees.");
      end if;
      Ast2Cfg.Post_Trans_Processing.Refine_Loops(World);
      Ast2Cfg.Post_Trans_Processing.Connect_Gotos(World); -- has to be called after loop refinement!
      Ast2Cfg.Post_Trans_Processing.Connect_Returns(World); -- has to be called after loop refinement!
      if Ast2Cfg.Transformation.Interprocedural then
         Ast2Cfg.Post_Trans_Processing.Link_Call_Nodes(World);
      else
         Put_Line(Debug, "[Transformation] Interprocedural option not set: Skipped linking the call nodes.");
      end if;
      Ast2Cfg.Post_Trans_Processing.Compute_Prefixes(World);
      Put_Line(Debug, "[Transformation] Post Transformation Processing - End.");
   end Post_Trans_Process;

   -- ** E L E M E N T   T R A N S F O R M A T I O N   H A N D L E R S

   -- Pre OP
   procedure Handle_Pre
     (Element :        Asis.Element;
      Control : in out Asis.Traverse_Control;
      State   : in out Actuals_For_Traversing.Traversal_State) is
      --
      Element_Kind:          Asis.Element_Kinds;
      Clause_Kind:           Asis.Clause_Kinds;
      Association_Kind:      Asis.Association_Kinds;
      Definition_Kind:       Asis.Definition_Kinds;
      Declaration_Kind:      Asis.Declaration_Kinds;
      Expression_Kind:       Asis.Expression_Kinds;
      Path_Kind:             Asis.Path_Kinds;
      Statement_Kind:        Asis.Statement_Kinds;
      --
      Flow_Ptr:              Flow_Types.Flow_Class_Ptr                       := null;
      Pkg_Ptr:               Pkgs.Pkg_Class_Ptr                              := null;
      Cfg_Ptr, New_Cfg_Ptr:  CFGs.CFG_Class_Ptr                              := null;
      --
      New_Node_Ptr, Prev_Node_Ptr, Ei_Node_Ptr, Call_Node_Ptr, Ret_Node_Ptr,
        Jump_Node_Ptr, Last_Node_Ptr, Param_Node_Ptr, Aux_Param_Node_Ptr,
        Loop_Node_Ptr, If_Node_Ptr:
        Nodes.Node_Class_Ptr                                                 := null;
      --
      Form_Decl:             Flow_Types.Formal_Decl_Class_Ptr                := null;
      Local_Alias:           Alias_Class_Ptr                                 := null;
      Inst_Alias:            Instance_Alias_Class_Ptr                        := null;
      Ref_Ptr:               Flow_Types.Ref_Class_Ptr                        := null;
   begin
      -- Print the Implementation Status of the Current Element Type
      Impl_Msg.Display_Message(" * [PRE ] ",Element);

      -- Save the current Element on the Stack
      Open_Element(Element);

      -- Identify the current Structures
      Get_Last_Flow_Info(Flow_Ptr, Pkg_Ptr, Cfg_Ptr);

      -- Check the Element Kind
      Element_Kind := Asis.Elements.Element_Kind (Element);
      case Element_Kind is

         when A_Clause =>

            -- C L A U S E S

            Clause_Kind := Asis.Elements.Clause_Kind (Element);
            case Clause_Kind is

               when A_With_Clause | A_Use_Package_Clause | A_Use_Type_Clause =>

                  -- Info: ** Entity Visibility & Inclusion Clauses (PRE)
                  --
                  --       These Visibility Clauses are the most common ones; They specify which
                  --       Entities are visible in the current Unit; For each begun Clause we
                  --       will add a new Reference Structure to the Transformation Data
                  --       Container; The Name of the Reference will be added in the HCHF Method;

                  if (Clause_Kind = A_With_Clause) then
                     Ref_Ptr := new Flow_Types.With_Ref;
                  elsif (Clause_Kind = A_Use_Package_Clause) then
                     Ref_Ptr := new Flow_Types.Use_Pkg_Ref;
                  elsif (Clause_Kind = A_Use_Type_Clause) then
                     Ref_Ptr := new Flow_Types.Use_Type_Ref;
                  else
                     Ref_Ptr := null;
                     Put_Line (Error, "[Transformation] [ERR] Unreachable Assertion failed!");
                  end if;

                  if Ref_Ptr /= null then
                     Ref_Class_Ptr_List.Append(Trans_Data.Refs, Ref_Ptr);
                  end if;

                  UStr_List.Clear(Trans_Data.Names);

               when others =>
                  null;
            end case;

         when An_Association =>

            -- A S S O C I A T I O N S

            Association_Kind := Asis.Elements.Association_Kind (Element);
            case Association_Kind is

               when A_Parameter_Association | A_Discriminant_Association |
                 An_Array_Component_Association | A_Record_Component_Association =>

                  -- Info: ** Parameter (-Class) Associations (PRE)
                  --
                  --       This is the PreOp for the Root-Node of a Parameter-Class Association
                  --       Subtree; PA Nodes are being used in the AST to hold the Values of
                  --       the Parameters in Subprogram Calls, for e.g. "j" in an "f(j)" Call;
                  --       They may also hold the Name of the Parameter itself in Calls like
                  --       "f(i => j)";
                  --
                  --       Note: The Discriminants are the Parameters of Tasks and share the
                  --       same Structure as the standard Parameters; They may also be used as
                  --       initial Input Parameters for Records;
                  --
                  --       Note: Array Component Associations are used to initialize Arrays,
                  --       like "Var_Array := (Var1, Var2, [..])"; They use only a Subset of the
                  --       Parameter-Class Association Characteristics and can be handled
                  --       together with the said Associations;
                  --
                  --       We must place a new Node (as Root for this Subtree) onto the
                  --       Parameter Stack; If this is an "i => j" Association then "i" will
                  --       be saved in Node.Name; "j" will be saved in Node.Info;

                  -- Create the Node
                  Param_Node_Ptr := new Nodes.Param_Node_Object;
                  Nodes.Init(Nodes.Param_Node_Object(Param_Node_Ptr.all), Flow_Ptr);
                  Nodes.Set_Info(Param_Node_Ptr.all, S2u("PARAM/NODE"));
                  -- Put it on the Parameter Stack
                  Open_Param_Non_Root(Param_Node_Ptr);
                  -- Make Room for the Name of the new Ids
                  UStr_List.Clear(Trans_Data.Names);

               when A_Generic_Association =>

                  -- Info: ** Generic Associations (PRE)
                  --
                  --       GAs are the Parameters of either Instantiations OR Formal
                  --       Declarations and hold the surrogate Types to be used in a generic
                  --       Entity; They have basically the same Structure as PAs, need however
                  --       more special Handling, since they need to be saved into either
                  --       the current Renaming/Instantiation Structure (for such Declarations)
                  --       OR into current Formal Declaration Structure (again, for such),
                  --       rather than in the Parameter Tree;
                  --       Since they cannot be nested, they don't need Stack Handling;

                  -- Create the Association Structure
                  Local_Alias := new Generic_Alias;
                  if ((Trans_Data.Rename /= null) and then
                      (Trans_Data.Rename.all in Flow_Types.Instance_Alias'Class)) then
                     -- The current Renaming Structure holds an Instantiation Record
                     -- Append the new Association to said Instantiation Record
                     Inst_Alias := Instance_Alias(Trans_Data.Rename.all)'Access;
                     Flow_Types.Alias_Class_Ptr_List.Append(Inst_Alias.Assocs, Local_Alias);
                  elsif (Flow_Types.Formal_Decl_Class_Ptr_List.Count(Trans_Data.Form_Decls) > 0) then
                     -- We have Formal Declarations
                     -- Append the new Association to the last (the current) Declaration
                     Form_Decl := Flow_Types.Formal_Decl_Class_Ptr_List.Get_Tail(Trans_Data.Form_Decls);
                     Flow_Types.Alias_Class_Ptr_List.Append(Form_Decl.Assocs, Local_Alias);
                  else
                     Put_Line(Warning,"[Transformation] [WARN] Dangling Generic Association!");
                  end if;
                  -- Make Room for the Name of the new Ids
                  UStr_List.Clear(Trans_Data.Names);

               when others =>
                  null;
            end case;

         when A_Definition =>

            -- D E F I N I T I O N S

            Definition_Kind := Asis.Elements.Definition_Kind (Element);
            case Definition_Kind is

               when A_Type_Definition | A_Formal_Type_Definition | A_Component_Definition |
                 A_Private_Type_Definition | A_Tagged_Private_Type_Definition |
                 A_Private_Extension_Definition =>

                  -- Info: ** (Formal) Type Definitons (PRE)
                  --
                  --       Note: The Considerations below apply to both Standard and Formal Types;
                  --
                  --       A Type Definition PreOp marks the Beginning of a -as the Name says-
                  --       Type Definition; The Type Definition specifies the Content of a Type;
                  --       It describes the Supertype, the Base Type, the Component and all the
                  --       used Identifiers;
                  --
                  --       This Section is mandatory for each *new* Type used in the Program;
                  --       This Type "Body" is being used both for Named (in Ordinary Type
                  --       Declarations, e.g. "type X is new Y") and Anonymous (in Variable
                  --       Declarations, e.g. "A: array() of Y") Types;
                  --
                  --       Since a Type Definition itself may hold Function Calls, we must
                  --       provide an adequate Parameter Tree;
                  --
                  --       If this is a Variable Declaration, we already have an open Parameter
                  --       Tree: the Initialization Tree for the Variable; In this Case we must
                  --       piggyback a new Tree on the Param Stack, on top of the open one; We
                  --       are not allowed to close the open one becuase we will need it later
                  --       during Initialization;
                  --
                  --       If this is a Type Declaration then we do not have an open Tree at all;
                  --
                  --       A Component Definition holds the Name of the Type of one Item in an
                  --       aggregate Type (Array, Record, ..); This Item may have a Name (e.g.
                  --       in Records as Part of a Component Declaration) or may be anonymous
                  --       (e.g. in Arrays);
                  --
                  --       In other Words: The Component Definition describes either the Base
                  --       Type of an Array, or the Type of a Record Component (NOTE: This Fact
                  --       is regarded as an Inconsistency; Even if the Record Components have
                  --       a semantical Meaning closer to that of the anonymous Array Components,
                  --       their Syntax however is similar to that of the Variables [Remember
                  --       that the Type of a Variable is regarded as a Supertype for the
                  --       Variable itself!]);

                  -- Create the Node
                  Param_Node_Ptr := new Nodes.Param_Node_Object;
                  Nodes.Init(Nodes.Param_Node_Object(Param_Node_Ptr.all), Flow_Ptr);
                  Nodes.Set_Info(Param_Node_Ptr.all, S2u("TYPE/DESCR"));
                  if Trans_Data.Decl /= null then
                     if (Trans_Data.Decl.VType.Descr_Tree /= null) then
                        Put_Line(Warning, "[Transformation] [WARN] " &
                            "Parameter tree clash in variable declaration! A tree already exists for this variable! Possible data loss.");
                     else
                        Trans_Data.Decl.VType.Descr_Tree :=
                            Nodes.Param_Node_Object(Param_Node_Ptr.all)'Access;
                     end if;
                  elsif not Formal_Decl_Class_Ptr_List.Empty(Trans_Data.Form_Decls) then
                     Formal_Decl_Class_Ptr_List.Get_Tail(Trans_Data.Form_Decls).VType.Descr_Tree :=
                       Nodes.Param_Node_Object(Param_Node_Ptr.all)'Access;
                  elsif Trans_Data.Type_Decl /= null then
                     if (Trans_Data.Type_Decl.Descr_Tree /= null) then
                        Put_Line(Warning, "[Transformation] [WARN] " &
                            "Parameter tree clash in type description! A tree already exists for this type! Possible data loss.");
                     else
                        Trans_Data.Type_Decl.Descr_Tree :=
                           Nodes.Param_Node_Object(Param_Node_Ptr.all)'Access;
                     end if;

                  else
                     Put_Line(Warning,"[Transformation] [WARN]" &
                              " No Structure available claims this Type Definition!");
                  end if;

                  -- Put it on the Parameter Stack
                  Open_Param_Root(Param_Node_Ptr);

               when A_Subtype_Indication =>

                  -- Info: ** Subtype Indications (PRE)
                  --
                  --       A Subtype Indication Node tells us that the current Entity being
                  --       defined (be it Type or Variable) inherits all the Attributes of a
                  --       given Type; It can be used in Variable Declarations (like Var:Type)
                  --       or int Type Declarations (like Type_New is new Type_Old);
                  --
                  --       The same way as with Type Definitions we should already have all the
                  --       Structures we need at this Point;

                  UStr_List.Clear(Trans_Data.Names);

               when A_Record_Definition =>

                  -- Info: ** Record Definitions (PRE)
                  --
                  --       Record Definitions are the Container Structures for Component
                  --       Declarations; We cannot influence the Handling of the Components here
                  --       because each Component has to decide for itself where it has to be
                  --       saved: note that Protected Objecs also serve as Containers for
                  --       Component Declarations;
                  --
                  --       Since we have a special Declarations List for the new Type (where each
                  --       Component Declaration will store itself) there is no need to create
                  --       any special Container Structure here;

                  null;

               when A_Discrete_Subtype_Definition =>

                  -- Info: ** Discrete Subtype Definitions (PRE)
                  --
                  --       This is a simple Type indicating a Range, like (1..2); It doesn't need
                  --       special Attention; Any Identifiers used will be saved automatically;
                  --

                  null;

               when A_Task_Definition =>

                  -- Info: ** Task Definitions (PRE)
                  --
                  --       The Name of this Structure is misleading; A TaskDef is actually a
                  --       Container for the Entry Declarations inside a Task
                  --       (*NON-Body*)Declaration; e.g. The Spec-Declaration of a Task T with
                  --       an Entry E will produce an AST with a Partial Structure like
                  --       "Task Declaration -> Task Definition -> Entry Declaration -> Expression";
                  --
                  --       We do not save the Entries we find in the Spec because we save them
                  --       when processing the Bodies, so there is nothing to be done about these
                  --       Task Definitions here either;

                  null;

               when others =>
                  null;
            end case;

         when A_Declaration =>

            -- D E C L A R A T I O N S

            -- Info: ** Declarations (PRE)
            --
            --       Declarations are independent Structures that need an empty Name List;

            UStr_List.Clear(Trans_Data.Names);

            Declaration_Kind := Asis.Elements.Declaration_Kind (Element);
            case Declaration_Kind is

               when An_Ordinary_Type_Declaration | A_Subtype_Declaration |
                 A_Private_Type_Declaration | A_Private_Extension_Declaration =>

                  -- Info: ** (Sub-)Type Declarations (PRE)
                  --
                  --       A Type Declaration creates a new *named* Type in the Program; It will
                  --       have, among others, a Name and a Type Definition (the Body);
                  --
                  --       When beginning a Type Declaration we must create the appropriate
                  --       Data Structure in the Transformation Data Container;

                  Trans_Data.Type_Decl := new Flow_Types.Type_Descr;
                  Trans_Data.Type_Decl.Type_Name := S2u("");
                  UStr_List.Init (Trans_Data.Type_Decl.Super_Type);
                  UStr_List.Init (Trans_Data.Type_Decl.Component_Type);
                  UStr_List_List.Init (Trans_Data.Type_Decl.Used_Idents);

                  if ((Declaration_Kind = A_Private_Type_Declaration) or
                      (Declaration_Kind = A_Private_Extension_Declaration)) then
                     Trans_Data.Type_Decl.Is_Private := True;
                  end if;

               when A_Package_Declaration | A_Package_Body_Declaration |
                 A_Protected_Type_Declaration | A_Single_Protected_Declaration |
                 A_Protected_Body_Declaration =>

                  -- Info: ** Flow Package Structure Declarations (PRE)
                  --
                  --       These are AST Structures that will result in Flow Packages in the
                  --       Flow World; We cannot do much about them at this Point since we know
                  --       nothing about them; Only when the first Child (containing their Name)
                  --       has been processed can we create the corresponding Flow Structures;

                  null;

               when A_Variable_Declaration | A_Component_Declaration |
                 A_Parameter_Specification | An_Entry_Index_Specification |
                 A_Discriminant_Specification | A_Loop_Parameter_Specification |
                 A_Constant_Declaration | A_Deferred_Constant_Declaration =>

                  -- Info: ** Variable-Class Declarations (PRE)
                  --
                  --       These Types all share a common Grammar containing Identifier(s)
                  --       (Variable Names, Parameter Names, ..), a Type and possibly an
                  --       initialization Expression;
                  --
                  --       When beginning with such a Declaration, we must initialize the
                  --       Declaration-Structure; It will hold the Information (temporarily)
                  --       until this Subtree has finished; We also have to initialize a Node
                  --       (Root) for the Parameter Tree to store the eventual Initialization;
                  --
                  --       An Exception is the Loop Parameter Specification; This is a Type that
                  --       occurs in a Statement and may interfere with the RHS-Tree of said
                  --       Statement, if we try to create an Init Tree for it; We will not create
                  --       any Init Tree for it, and any Function Calls will be appended to the
                  --       RHS Tree;
                  --
                  --       When "Ignore Mode" is turned on, we do not care about any of these
                  --       Declarations; This Mode is usually on when we are dealing with
                  --       Function Declarations and we need to ignore their Parameters;
                  --       A Type Declaration for a Subprogram Pointer may contain Parameters
                  --       in Non-Ignore Mode - Parameter that we should actually ignore - so we
                  --       must also check in Non-Ignore Mode if we have any Type Structures
                  --       open; Should this be the Case, we must then deliberately skip this
                  --       Declaration;

                  if ((not Trans_Data.Ignore_Mode) and
                      (not ((Declaration_Kind = A_Parameter_Specification) and
                            (Trans_Data.Type_Decl /= null)))) then

                     -- Initialize the Declaration Structure
                     Trans_Data.Decl := new Flow_Types.Declaration;
                     UStr_List.Init(Trans_Data.Decl.Vars);
                     Flow_Types.Clear(Trans_Data.Decl.VType);

                     -- Is it private?
                     if Declaration_Kind = A_Deferred_Constant_Declaration then
                        Trans_Data.Decl.VType.Is_Private := True;
                     end if;

                     -- Create the Param Node
                     Param_Node_Ptr := new Nodes.Param_Node_Object;
                     Nodes.Init(Nodes.Param_Node_Object(Param_Node_Ptr.all), Flow_Ptr);
                     Nodes.Set_Info(Param_Node_Ptr.all, S2u("INIT/NODE"));

                     -- Save the Root of the Param Tree
                     Trans_Data.Decl.Init := Flow_Types.Flow_Object(Param_Node_Ptr.all)'Access;

                     -- Put it on the Parameter Stack
                     if (Declaration_Kind /= A_Loop_Parameter_Specification) then
                        Open_Param_Root(Param_Node_Ptr);
                     end if;

                  end if;

               when A_Procedure_Declaration | A_Function_Declaration |
                 An_Entry_Declaration |
                 A_Procedure_Body_Stub | A_Function_Body_Stub | A_Task_Body_Stub =>

                  -- Info: ** CFG-Class Spec Declarations (PRE)
                  --
                  --       These are basically the Asis Spec Types that have Parameters,
                  --       Parameters that are NOT being taken into Consideration by the
                  --       Transformation; However, since the Processing of Arguments is done
                  --       regardless of the Context, we must instruct the Transformation Logic
                  --       to skip any Parameter Declarations found here;
                  --
                  --       The Parameter WILL be added when Processing the Body;
                  --       We will reset the Ignore_Mode when these Structures have finished;

                  Trans_Data.Ignore_Mode := True;

               when A_Procedure_Body_Declaration | A_Function_Body_Declaration |
                 An_Entry_Body_Declaration |
                 A_Single_Task_Declaration | A_Task_Type_Declaration | A_Task_Body_Declaration =>

                  -- Info: ** CFG-Class Body Declarations (PRE)
                  --
                  --       These ASIS Types result in CFGs in the Flow World; Since we'll only
                  --       know their Name after their first Child has finished we cannot create
                  --       them now;

                  null;

               when A_Generic_Package_Declaration |
                 A_Generic_Procedure_Declaration | A_Generic_Function_Declaration =>

                  -- Info: ** Generic Structures (PRE)
                  --
                  --       A generic Entity begins; We will only be able to create the
                  --       corresponding Flow Structure after the first Child (with the Name)
                  --       has finished; Since we do not have the Structure yet we will keep the
                  --       formal Declarations (if any) in the Transformation Data Container
                  --       for the Time being;

                  Flow_Types.Formal_Decl_Class_Ptr_List.Clear(Trans_Data.Form_Decls);

               when A_Formal_Object_Declaration | A_Formal_Type_Declaration |
                 A_Formal_Procedure_Declaration | A_Formal_Function_Declaration |
                 A_Formal_Package_Declaration | A_Formal_Package_Declaration_With_Box =>

                  -- Info: ** Formal Declarations (PRE)
                  --
                  --       We are beginning here with a formal Declaration; during the Process
                  --       we don't need Parameters; we need however a Structure for the formal
                  --       Declaration itself; All the formal Declarations will be stored in a
                  --       List, until the Generic Flow Object has been created; Note that said
                  --       Object does not exist yet (since we don't have its Name);
                  --
                  --       We also have to initialize a Node (Root) for the Parameter Tree to
                  --       store the eventual Initialization;

                  -- Ignore eventual Parameter Declarations
                  Trans_Data.Ignore_Mode := True;

                  -- Create the correct formal Declaration Object
                  case Declaration_Kind is
                     when A_Formal_Object_Declaration =>
                        Form_Decl := new Obj_Formal_Decl;
                     when A_Formal_Type_Declaration =>
                        Form_Decl := new Type_Formal_Decl;
                     when A_Formal_Procedure_Declaration =>
                        Form_Decl := new Proc_Formal_Decl;
                     when A_Formal_Function_Declaration =>
                        Form_Decl := new Func_Formal_Decl;
                     when A_Formal_Package_Declaration | A_Formal_Package_Declaration_With_Box =>
                        Form_Decl := new Pkg_Formal_Decl;
                     when others =>
                        Put_Line(Error, "[Transformation] Unreachable assertion failed!");
                        Form_Decl := null;
                  end case;
                  if (Form_Decl /= null) then
                     -- Initialize the Structure
                     UStr_List.Init(Form_Decl.Vars);
                     Flow_Types.Clear(Form_Decl.VType);
                     Form_Decl.Init := null;
                     -- Append the Structure to the List of formal Declarations for this Entity
                     Flow_Types.Formal_Decl_Class_Ptr_List.Append(Trans_Data.Form_Decls, Form_Decl);
                  end if;

                  -- Create the Param Node
                  Param_Node_Ptr := new Nodes.Param_Node_Object;
                  Nodes.Init(Nodes.Param_Node_Object(Param_Node_Ptr.all), Flow_Ptr);
                  Nodes.Set_Info(Param_Node_Ptr.all, S2u("INIT/NODE"));

                  -- Save the Root of the Param Tree
                  Form_Decl.Init := Flow_Types.Flow_Object(Param_Node_Ptr.all)'Access;

                  -- Put it on the Parameter Stack
                  Open_Param_Root(Param_Node_Ptr);

               when A_Procedure_Renaming_Declaration | A_Function_Renaming_Declaration |
                 A_Generic_Procedure_Renaming_Declaration | A_Generic_Function_Renaming_Declaration |
                 A_Package_Renaming_Declaration | A_Generic_Package_Renaming_Declaration |
                 An_Object_Renaming_Declaration | An_Exception_Renaming_Declaration  =>

                  -- Info: ** Entity Renaming Declarations (PRE)
                  --
                  --       These are the ASIS Renaming Declarations; We don't need any Parameters
                  --       with them, we only need their Names; We must set up a Structure to
                  --       hold the new Name Association until we can save it into the current
                  --       Flow Object;

                  -- Ignore eventual Parameter Declarations
                  Trans_Data.Ignore_Mode := True;

                  -- Set up the Data Structure
                  if ((Declaration_Kind = A_Procedure_Renaming_Declaration) or
                      (Declaration_Kind = A_Generic_Procedure_Renaming_Declaration)) then
                     Trans_Data.Rename := new Proc_Alias;
                  elsif ((Declaration_Kind = A_Function_Renaming_Declaration) or
                         (Declaration_Kind = A_Generic_Function_Renaming_Declaration)) then
                     Trans_Data.Rename := new Func_Alias;
                  elsif ((Declaration_Kind = A_Package_Renaming_Declaration) or
                         (Declaration_Kind = A_Generic_Package_Renaming_Declaration)) then
                     Trans_Data.Rename := new Pkg_Alias;
                  elsif (Declaration_Kind = An_Object_Renaming_Declaration) then
                     Trans_Data.Rename := new Obj_Alias;
                  elsif (Declaration_Kind = An_Exception_Renaming_Declaration) then
                     Trans_Data.Rename := new Except_Alias;
                  end if;

                  -- Create a new Param Node
                  Param_Node_Ptr := new Nodes.Param_Node_Object;
                  Nodes.Init(Nodes.Param_Node_Object(Param_Node_Ptr.all), Flow_Ptr);
                  Nodes.Set_Info(Param_Node_Ptr.all, S2u("RENAME/TREE"));

                  -- Save the Root of the Param Tree
                  Trans_Data.Rename.PTree := Flow_Types.Flow_Object(Param_Node_Ptr.all)'Access;

                  -- Put it on the Parameter Stack
                  Open_Param_Root(Param_Node_Ptr);

               when A_Package_Instantiation |
                 A_Procedure_Instantiation | A_Function_Instantiation =>

                  -- Info: ** Generic Instantiations (PRE)
                  --
                  --       These are Instantiations of Generic Types; They share a common Grammar
                  --       with the Renamings to a certain Extent, have however Generic
                  --       Associations; We will use an Alias derived Structure, i.e.
                  --       an Instantiation Alias for them, that also has Room for the said GAs;
                  --
                  --       Note that the Generic Associations of this Instantiation will be added
                  --       one by one when PreOp-ing the GAs;

                  if (Declaration_Kind = A_Package_Instantiation) then
                     Trans_Data.Rename := new Pkg_Inst_Alias;
                  elsif (Declaration_Kind = A_Procedure_Instantiation) then
                     Trans_Data.Rename := new Proc_Inst_Alias;
                  elsif (Declaration_Kind = A_Function_Instantiation) then
                     Trans_Data.Rename := new Func_Inst_Alias;
                  end if;

               when others =>
                  null;
            end case;

         when An_Expression =>

            -- E X P R E S S I O N S

            Expression_Kind := Asis.Elements.Expression_Kind (Element);
            case Expression_Kind is

               when A_Positional_Array_Aggregate | A_Record_Aggregate =>

                  -- Info: ** Aggregates (PRE)
                  --
                  --       These Structures are used for Array/Record/* Initializations, like
                  --       " := (a,b,c)"; We will create an extra Node on the Parameter Stack
                  --       to wrap all Children together;

                  -- Create the Node
                  Param_Node_Ptr := new Nodes.Param_Node_Object;
                  Nodes.Init(Nodes.Param_Node_Object(Param_Node_Ptr.all), Flow_Ptr);
                  if (Expression_Kind = A_Positional_Array_Aggregate) then
                     Nodes.Set_Info(Param_Node_Ptr.all, S2u("ARRAY/AGGREGATE"));
                  elsif (Expression_Kind = A_Record_Aggregate) then
                     Nodes.Set_Info(Param_Node_Ptr.all, S2u("REC/AGGREGATE"));
                  end if;

                  -- Put it on the Parameter Stack
                  Open_Param_Non_Root(Param_Node_Ptr);

               when An_Attribute_Reference =>

                  -- Info: ** Attribute References (PRE)
                  --
                  --       ARs are "Var'Attribute" Type Constructs; No PreOp is needed;

                  null;

               when An_Allocation_From_Subtype =>

                  -- Info: ** Allocation From Subtype (PRE)
                  --
                  --       The Subtree beginning here describes the Type of the Object being
                  --       instantiated (e.g. "Var := new Type");
                  --
                  --       The complicated Part is that such Instantiations may occur just about
                  --       anywhere in the Program; In order not to conflict with Type
                  --       Definitions (that have a similar Structure) AND because Allocations
                  --       From Subtype are being used in a similar Way to Function Calls, we
                  --       will spawn a Parameter Node for each such Allocation; This Node will
                  --       hold the Name of the Type used, and -if necessary- further Parameters;

                  -- Create the Node
                  Param_Node_Ptr := new Nodes.Param_Alloc_Node_Object;
                  Nodes.Init(Nodes.Param_Alloc_Node_Object(Param_Node_Ptr.all), Flow_Ptr);
                  Nodes.Set_Info(Param_Node_Ptr.all, S2u("ALLOC/NODE"));
                  -- Put it on the Parameter Stack
                  Open_Param_Non_Root(Param_Node_Ptr);
                  -- Make Room for the Name of the new Ids
                  UStr_List.Clear(Trans_Data.Names);

               when A_Type_Conversion =>

                  -- Info: ** Type Conversions (PRE)
                  --
                  --       The Type Conversion is the typical Type Casting; There is no
                  --       Pre-Operation we can do here; The Work is done in HCHF;

                  null;

               when An_Identifier =>

                  -- Info: ** Identifiers (PRE)
                  --
                  --       At this Point there is nothing we can do about Identifiers; They
                  --       are generally added to the Trans_Data.Names List for further
                  --       Processing; At this Point however we don't even know their Names;

                  null;

               when An_Operator_Symbol =>

                  -- Info: ** Operator Symbols (PRE)
                  --
                  --       Since Operator Symbols are usually used as Names for Functions we will
                  --       save them into the Trans_Data.Names List for further Processing;
                  --       For now, there is nothing we can do;

                  null;

               when A_Function_Call =>

                  -- Info: ** Function Calls (PRE)
                  --
                  --       This is a genuine Function Call; Functions ARE Parameters and also
                  --       HAVE Parameters themselves, thus we must update the Parameter Stack;
                  --       First we create a Param Node for the Function itself, that we will
                  --       place on top of the current Parameter Stack;
                  --       All Parameters found later (deeper in the Hierarchy) will be saved as
                  --       Children of this Parameter Call Node;

                  -- Create the Node
                  Param_Node_Ptr := new Nodes.Param_Call_Node_Object;
                  Nodes.Init(Param_Node_Ptr.all, Flow_Ptr, Element);
                  Nodes.Set_Info(Param_Node_Ptr.all, S2u("FUNC/PARAM/ROOT"));

                  -- Set the functions target
                  Nodes.Set_Call_Target(Nodes.Call_Node_Object(Param_Node_Ptr.all),
                                        Asis.Expressions.Corresponding_Called_Function(Element));

                  -- Put it on the Parameter Stack
                  Open_Param_Non_Root(Param_Node_Ptr);
                  -- Make Room for the Name of the called Function
                  UStr_List.Clear(Trans_Data.Names);

               when A_Selected_Component =>

                  -- Info: ** Selected Components (PRE)
                  --
                  --       SCs are Items of a Name1.Name2.Name3.. Address; Since each of the
                  --       found Names will be saved in the Trans_Data.Names List, we will be
                  --       able to reconstruct the complete Name; At this Point however there
                  --       is nothing we can do about them;

                  null;

               when An_Indexed_Component | A_Slice =>

                  -- Info: ** Indexed (-Class) Components (PRE)
                  --
                  --       ICs are all the Structures that have an Index (e.g. A(3)); Since they
                  --       cannot be completely analysed (due to thir extreme possible
                  --       Complexity), we will only try to save their Name and some Information
                  --       about the Index;
                  --
                  --       For now, we create a Parameter Node to hold any complex Expressions
                  --       that may have been used;

                  -- Create the Node
                  Param_Node_Ptr := new Nodes.Param_Node_Object;
                  Nodes.Init(Param_Node_Ptr.all, Flow_Ptr, Element);
                  if Expression_Kind = An_Indexed_Component then
                     Nodes.Set_Info(Param_Node_Ptr.all, S2u("IDX/COMP"));
                  else
                     Nodes.Set_Info(Param_Node_Ptr.all, S2u("SLICE"));
                  end if;

                  -- Put it on the Parameter Stack
                  Open_Param_Non_Root(Param_Node_Ptr);

                  -- Make Room for the Name of the Component
                  UStr_List.Clear(Trans_Data.Names);

               when others =>
                  null;
            end case;

         when A_Statement | A_Path =>

            -- S T A T E M E N T S / P A T H S

            -- ** GENERAL STATEMENT/PATH (PRE) HANDLING

            -- Info: ** Statements and Paths (PRE)
            --
            --       Each Statement (imperative LOC) in Ada produces one Statement Node in the
            --       AST; Complex Statements begin here and have associated Subtrees for each of
            --       their Components; ASIS also generates for each possible Branch in an "if"
            --       or "case" Alternative a so called "Path"-Node; One Path Node marks the
            --       beginning of a possible Alternative;
            --
            --       The standard Procedure when dealing with Statements/Paths is to create a
            --       general Purpose Node and put it on the Node Stack of the current Flow;
            --       Special Statements however require special Nodes, so, at a later Point in
            --       Time, we may decide to replace the general Node with a more specialized one;
            --
            --       Statements usually occur only in CFGs, thus, in most of the Cases we will
            --       have already created a CFG before processing a Statement;
            --       However, Statements may also appear before creating a CFG - when processing
            --       the Initialization Part of a Package; In such Cases we will create an
            --       Init-CFG and append it to the current Package;

            -- Info: ** Node Stack Usage
            --
            --       When processing a CFG we also have a Node Stack Structure for it; The Top of
            --       the Stack is always the current Statement being processed; When a new
            --       Statement begins, we REMOVE the old one from the Stack; This would mean the
            --       Stack keeps a Height of one (1) over the entire Execution of the Program;
            --
            --       However, the Stack DOES grow when we encounter Nodes that we will need at a
            --       later Point in Time again; The two most important such Cases are the
            --       Branching-Class and the Loop-Class Nodes;
            --
            --       In the BC we have the "If" and the "EndIf" Nodes that we will need over and
            --       over again to make the correct Links after a Branch finishes;
            --       In the LC we have the Loop Head that we will need again, after all Nodes in
            --       Loop itself have finished;

            -- Check for the Parent CFG
            if (Cfg_Ptr = null) then
               -- No CFG - this is the executable Part of a Package; look for the parent Package
               if (Pkg_Ptr = null) then
                  -- No Parent at all!
                  Put_Line (Error, "[Transformation] [ERR] No parent Package for Init CFG!");
                  return;
               else
                  -- A Parent Pkg exists -> create & initialize the Init CFG
                  New_Cfg_Ptr := new CFGs.Init_Object;
                  CFGs.Init(New_Cfg_Ptr.all, S2u("Init_CFG"), Flow_Ptr, Element);
                  -- Update the Succs of the Father Package
                  Pkgs.Add_Succ(Pkg_Ptr.all, Flow_Types.Flow_Object(New_Cfg_Ptr.all)'Access);
                  Put_Line (Debug, "[Transformation] New CFG: " & CFGs.Get_Name(New_Cfg_Ptr.all));
                  -- Put the new CFG on the Flow-Stack & update Variables
                  Open_Flow(Flow_Types.Flow_Class_Ptr(New_Cfg_Ptr), Get_Last_Flow_Elem);
                  Get_Last_Flow_Info(Flow_Ptr, Pkg_Ptr, Cfg_Ptr);
               end if;
            end if;

            -- Info: ** Identifier Policy when (PRE) processing Statements
            --
            --       Since the Complexity of the AST makes it difficult to analyse completely,
            --       we will save per Default all Identifiers found when processing a Statement
            --       into the Statement itself; This includes Variables, Subprogram Names, ..
            --       In the PreOp we will clear the Identifier List and in the PostOp we will
            --       save the found Names;

            -- Make Room for more Identifiers
            UStr_List.Clear(Trans_Data.Names);

            -- Create the Node, set the Node Information and save it
            New_Node_Ptr := new Nodes.Node_Object;
            Nodes.Init(New_Node_Ptr.all, Flow_Ptr, Element);
            Nodes.Set_Info(New_Node_Ptr.all, S2u(First_Line(Element)));
            -- Replace the Top of the Node Stack with the New Node
            Prev_Node_Ptr := Replace_Last_Node(New_Node_Ptr);

            -- Info: ** Parameter Tree (PRE) Handling
            --
            --       Each Statement may contain Function Calls that require a Parameter Tree;
            --       This Parameter Tree will be the RHS (Right Hand Side) Tree; Assignment
            --       Statement may also have a LHS Tree for the Left Side;
            --
            --       In Conclusion: When beginning the Processing of a Statement we must also
            --       create an empty Param Tree (the Default for a general Statement);

            -- Purge any Lefotovers from the current Param Tree
            Close_All_Params;

            -- Create the Root of the new Param Tree
            Param_Node_Ptr := new Nodes.Param_Node_Object;
            Nodes.Init(Nodes.Param_Node_Object(Param_Node_Ptr.all), Flow_Ptr);
            Nodes.Set_Info(Param_Node_Ptr.all, S2u("STMT/PAR"));
            -- Put the new Root Node on the Parameter Stack
            Open_Param_Root(Param_Node_Ptr);

            -- ** SPECIAL PATH (PRE) HANDLING

            -- Info: ** Paths (PRE)
            --
            --       Path Nodes are on Top of an "If"/"Case"/"Select*" Branch; We could ignore
            --       them, since they contain no Information; However, we will save them for a
            --       better Readabilty of the CFG on the Screen;

            if Element_Kind = A_Path then
               Path_Kind := Asis.Elements.Path_Kind (Element);
               case Path_Kind is
                  when An_If_Path | An_Elsif_Path | A_Case_Path =>
                     -- Make no Changes to the Node; The Information should already contain
                     -- the Condition of the Alternatives;
                     null;
                  when A_Select_Path =>
                     Nodes.Set_Info(New_Node_Ptr.all, S2u("select"));
                  when An_Or_Path =>
                     Nodes.Set_Info(New_Node_Ptr.all, S2u("or"));
                  when An_Else_Path =>
                     Nodes.Set_Info(New_Node_Ptr.all, S2u("else"));
                  when A_Then_Abort_Path =>

                     -- Info: ** Abort CFG
                     --
                     --       The Code in an "Abort" Path must be packed into a special
                     --       Abort-CFG; The Link between the Call Node and the CFG we will
                     --       create, will be Id of the Call Node; We will save this Id into
                     --       the Link Name Variable (LName) of the new CFG; With this Link
                     --       Name, we will be able to identify which Node calls which CFG;

                     -- Create a new Abort Node (Call Node since Jump Point into the Abort CFG)
                     Call_Node_Ptr := new Nodes.Abort_Node_Object;
                     Nodes.Init(Call_Node_Ptr.all, Flow_Ptr, Element);
                     Nodes.Set_Info(Call_Node_Ptr.all, S2u("abort"));
                     Last_Node_Ptr := Replace_Last_Node_And_Relink(Call_Node_Ptr, Prev_Node_Ptr);
                     -- Update the Variables
                     New_Node_Ptr := Call_Node_Ptr;

                     -- Spawn a new Abort CFG
                     New_Cfg_Ptr := new CFGs.Abort_Object;
                     CFGs.Init(New_Cfg_Ptr.all, S2u("Abort_CFG"), Flow_Ptr, Element);
                     Put_Line (Debug, "[Transformation] New CFG: " & CFGs.Get_Name(New_Cfg_Ptr.all));
                     -- Make the correct Links
                     CFGs.Add_Succ(Cfg_Ptr.all, Flow_Types.Flow_Class_Ptr(New_Cfg_Ptr));
                     -- Set the Name - generated from the CFG's Id
                     CFGs.Set_LName(New_Cfg_Ptr.all, S2u(Id.To_CFG_Name(CFGs.Get_Id(New_Cfg_Ptr.all))));
                     -- Update the newly created Call Node
                     Nodes.Add_Call_Item (Nodes.Abort_Node_Object(Call_Node_Ptr.all),
                                          S2u(CFGs.Get_LName(New_Cfg_Ptr.all)));
                     -- Push the new CFG onto the Flow-Stack
                     Open_Flow(Flow_Types.Flow_Class_Ptr(New_Cfg_Ptr), Element);
                     Get_Last_Flow_Info(Flow_Ptr, Pkg_Ptr, Cfg_Ptr);
                     New_Node_Ptr := null;

                  when Not_A_Path =>
                     null;
               end case;

               -- Set the RHS Tree for the Path Node
               if (Param_Node_Ptr /= null) and (New_Node_Ptr /= null) then
                  Nodes.Set_RHS_Tree(New_Node_Ptr.all, Get_Last_Param_Node);
               end if;

               -- Info: ** Finished Paths (PRE)
               --
               --       If this Element was a Path then we must terminate here; The Rest of the
               --       Handling is is for Statements only;

               return;

            end if; -- If "Path"

            -- ** SPECIAL STATEMENT (PRE) HANDLING

            -- Info: ** Special Statement Handling
            --
            --       We are only handling "special" Statements here; They all replace the general
            --       Purpose Node with a more specialized one;

            Statement_Kind := Asis.Elements.Statement_Kind (Element);
            case Statement_Kind is

               when An_Assignment_Statement =>

                  -- Info: ** Assignment Statements (PRE)
                  --
                  --       We are replacing the general Node with an Assignment one; The Param
                  --       Tree we have created earlier as RHS Tree will be used for this
                  --       Statement as LHS Tree; Later we will also create the RHS Tree for
                  --       this Assignment Statement;

                  -- Create and initialize the new Assignment Node
                  New_Node_Ptr := new Nodes.Assign_Node_Object;
                  Nodes.Init(New_Node_Ptr.all, Flow_Ptr, Element);
                  Nodes.Set_Info(New_Node_Ptr.all, S2u(First_Line(Element)));
                  Last_Node_Ptr := Replace_Last_Node_And_Relink(New_Node_Ptr, Prev_Node_Ptr);

                  -- Mutate the RHS Tree to LHS Tree
                  Nodes.Set_LHS_Tree(Nodes.Assign_Node_Object(New_Node_Ptr.all), Param_Node_Ptr);
                  Nodes.Set_Info(Get_Last_Param_Node.all, S2u("ASSIGN/LHS"));
                  Nodes.Set_RHS_Tree(New_Node_Ptr.all, null);

               when A_Goto_Statement | An_Exit_Statement =>

                  -- Info: ** Jump-Class Statements (PRE)
                  --
                  --       We replace the general Node with a Jump Node; for now, no Target
                  --       is being set since none known;

                  -- Create a new Jump Node
                  if Statement_Kind = A_Goto_Statement then
                     Jump_Node_Ptr := new Nodes.Goto_Jump_Node_Object;
                  else
                     -- This is an Exit Node; check its type: if it has a "with" - it's a
                     -- complex one, if not it's a trivial one
                     if (Asis.Elements.Expression_Kind(Asis.statements.Exit_Condition(Element))
                           = Not_An_Expression) then
                        -- This node does not have an exit condition -> trivial
                        Jump_Node_Ptr := new Nodes.Trivial_Exit_Jump_Node_Object;
                     else
                        -- This node does have an exit condition -> complex
                        Jump_Node_Ptr := new Nodes.Complex_Exit_Jump_Node_Object;
                     end if;
                  end if;
                  Nodes.Init(Jump_Node_Ptr.all, Flow_Ptr, Element);
                  Nodes.Set_Info(Jump_Node_Ptr.all, S2u(First_Line(Element)));
                  Last_Node_Ptr := Replace_Last_Node_And_Relink(Jump_Node_Ptr, Prev_Node_Ptr);
                  New_Node_Ptr := Jump_Node_Ptr;

               when A_Procedure_Call_Statement | A_Block_Statement | A_Raise_Statement |
                 An_Entry_Call_Statement | A_Requeue_Statement | A_Requeue_Statement_With_Abort |
                 An_Abort_Statement | An_Accept_Statement  =>

                  -- Info: ** Call-Class Statements Statements (PRE)
                  --
                  --       We replace the general Node with a Call Node; for now, no Calls are
                  --       being appended yet since none known;

                  -- Create a new Call Node
                  if (Statement_Kind = A_Procedure_Call_Statement) or
                    (Statement_Kind = A_Block_Statement) or
                    (Statement_Kind = A_Raise_Statement) then
                     Call_Node_Ptr := new Nodes.Call_Node_Object;
                  elsif (Statement_Kind = An_Entry_Call_Statement) or
                    (Statement_Kind = A_Requeue_Statement) or
                    (Statement_Kind = A_Requeue_Statement_With_Abort) then
                     Call_Node_Ptr := new Nodes.Entry_Call_Node_Object;
                  elsif (Statement_Kind = An_Abort_Statement) then
                     Call_Node_Ptr := new Nodes.Abort_Node_Object;
                  elsif (Statement_Kind = An_Accept_Statement) then
                     Call_Node_Ptr := new Nodes.Accept_Node_Object;
                  else
                     Put_Line(Error, "[Transformation] Unreachable assertion failed!");
                     return;
                  end if;

                  Nodes.Init(Call_Node_Ptr.all, Flow_Ptr, Element);
                  Nodes.Set_Info(Call_Node_Ptr.all, S2u(First_Line(Element)));
                  Last_Node_Ptr := Replace_Last_Node_And_Relink(Call_Node_Ptr, Prev_Node_Ptr);
                  New_Node_Ptr := Call_Node_Ptr;

                  -- If this was a procedure call, also save its target CFG
                  -- Note: You can add here the targets for other call nodes as well
                  if (Statement_Kind = A_Procedure_Call_Statement) then
                     Nodes.Set_Call_Target(Nodes.Call_Node_Object(Call_Node_Ptr.all),
                                           Asis.Statements.Corresponding_Called_Entity(Element));
                  end if;

                  if ((Statement_Kind = A_Block_Statement) or
                      (Statement_Kind = An_Accept_Statement)) then

                     -- Info: ** Block-Class Statements (PRE)
                     --
                     --       A Block Node is the Beginning Mark of a new CFG so we must also
                     --       create one;
                     --
                     --       The same goes for Accept Statements; They also mark the Beginning
                     --       of a new CFG, complete with Parameters, declarative Part, a.s.o.
                     --
                     --       The Link between the Call Node and the CFG we will create, will be
                     --       Id of the Call Node; We will save this Id into the Link Name
                     --       Variable (LName) of the new CFG; With this Link Name, we will be
                     --       able to identify which Node calls which CFG even if we change the
                     --       Name of the CFG on the way;
                     --
                     --       Note: For Accept Statement it is vital to Change said Name, so we
                     --       know which Entry is being accepted;

                     -- Spawn a new CFG
                     if (Statement_Kind = A_Block_Statement) then
                        New_Cfg_Ptr := new CFGs.Block_Object;
                        CFGs.Init(New_Cfg_Ptr.all, S2u("Block_CFG"), Flow_Ptr, Element);
                     else
                        New_Cfg_Ptr := new CFGs.Entry_Object;
                        CFGs.Init(New_Cfg_Ptr.all, S2u("Accept_CFG"), Flow_Ptr, Element);
                     end if;
                     -- Rename the CFG (with the Name used in the Call Node)
                     CFGs.Set_LName(New_Cfg_Ptr.all, S2u(Id.To_CFG_Name(CFGs.Get_Id(New_Cfg_Ptr.all))));
                     Put_Line (Debug, "[Transformation] New CFG: " & CFGs.Get_Name(New_Cfg_Ptr.all));
                     -- Make the correct Links
                     CFGs.Add_Succ(Cfg_Ptr.all, Flow_Types.Flow_Class_Ptr(New_Cfg_Ptr));
                     -- Update the newly created Call Node
                     Nodes.Add_Call_Item (Nodes.Call_Node_Object(Call_Node_Ptr.all),
                                          S2u(CFGs.Get_LName(New_Cfg_Ptr.all)));
                     Nodes.Set_Call_Destination (Nodes.Call_Node_Object(Call_Node_Ptr.all),
                                                 Flow_Class_Ptr(New_Cfg_Ptr));

                     -- Push the new CFG onto the Flow-Stack
                     Open_Flow(Flow_Types.Flow_Class_Ptr(New_Cfg_Ptr), Element);
                     -- Update Variables
                     Get_Last_Flow_Info(Flow_Ptr, Pkg_Ptr, Cfg_Ptr);
                     New_Node_Ptr := null;

                  else

                     -- Info: ** Parameter Tree Handling for Call Nodes
                     --
                     --       Note: The succs of the Nodes on the Parameter Stack represent the
                     --       Parameters of a Call, whereas the succ-Relationship in the CFG
                     --       represents the Control Flow Order;
                     --
                     --       At this Point we already have a Parameter Tree, BUT we need to put
                     --       the Call Node itself onto the Parameter Stack, because only then
                     --       will the PT-Successors of this Node symbolize the Subprogram
                     --       Parameter correctly; However, we can't put the Call Node directly
                     --       onto the Parameter Stack since this Node is part of the CFG;
                     --
                     --       We will create a Parameter Call Node for the PT as Avatar of the
                     --       Call Node in the CFG; This new Parameter Node will replace the old
                     --       one we created earlier;

                     -- Create the new Parameter Call Node
                     Param_Node_Ptr := new Nodes.Param_Call_Node_Object;
                     Nodes.Init(Param_Node_Ptr.all, Flow_Ptr, Element);
                     Nodes.Set_Info(Param_Node_Ptr.all, S2u("PROC/PARAM/ROOT"));
                     -- Put the new PN onto the Parameter Stack
                     Aux_Param_Node_Ptr := Get_Last_Param_Node;
                     Close_Last_Param;
                     Open_Param_Root(Param_Node_Ptr);

                     -- If this was a procedure call, also save its target CFG
                     -- Note: You can add here the targets for other call nodes as well
                     -- Note2: We duplicate here the call node for a procedure call statement; the
                     -- parameter node needs the target element just as well!
                     if (Statement_Kind = A_Procedure_Call_Statement) then
                        Nodes.Set_Call_Target(Nodes.Call_Node_Object(Param_Node_Ptr.all),
                                              Asis.Statements.Corresponding_Called_Entity(Element));
                     end if;

                     -- Remove the old Parameter Node
                     Nodes.Free(Aux_Param_Node_Ptr);
                  end if;

               when An_If_Statement | A_Case_Statement | A_Selective_Accept_Statement |
                 A_Conditional_Entry_Call_Statement | A_Timed_Entry_Call_Statement |
                 An_Asynchronous_Select_Statement =>

                  -- Info: ** Branch-Class Nodes (PRE)
                  --
                  --       Because the Statement Processing always replaces the top Node on the
                  --       Node Stack, AND because we cannot afford to lose Track of the first
                  --       Node in a Branching Statement (we want to be able to link the first
                  --       Node of the next Branch to the top Node as well), we must put this
                  --       first Node twice on the Stack together with a pseudo "EndIf" Node;
                  --
                  --       The Stack Top will become "If-EndIf-If"; The last "If" will get
                  --       replaced by the Nodes in the Branch; At the End of a Branch we will
                  --       link the last Node to "EndIf" and restore the "If-EndIf-If" Order;
                  --       This way, all Nodes situated in the Branches will be integrated
                  --       between the first "If" and the last "EndIf" Node;

                  -- Update the Node Info
                  if (Statement_Kind = A_Selective_Accept_Statement) then
                     Nodes.Set_Info(New_Node_Ptr.all, S2u("select/acc"));
                  elsif ((Statement_Kind = A_Conditional_Entry_Call_Statement) or
                         (Statement_Kind = A_Timed_Entry_Call_Statement)) then
                     Nodes.Set_Info(New_Node_Ptr.all, S2u("select/entry"));
                  elsif (Statement_Kind = An_Asynchronous_Select_Statement) then
                     Nodes.Set_Info(New_Node_Ptr.all, S2u("select/abort"));
                  else
                     If_Node_Ptr := new Nodes.Branch_Node_Object;
                     Nodes.Init(If_Node_Ptr.all, Flow_Ptr, Element);
                     Nodes.Set_Info(If_Node_Ptr.all, S2u("if/case"));
                     Last_Node_Ptr := Replace_Last_Node_And_Relink(If_Node_Ptr, Prev_Node_Ptr);
                     New_Node_Ptr := If_Node_Ptr;
                  end if;
                  -- Create the Pseudo-Node (the "EndIf/Case/Select")
                  Ei_Node_Ptr := new Nodes.Node_Object;
                  Nodes.Init(Ei_Node_Ptr.all, Flow_Ptr, Element);
                  if ((Statement_Kind = An_If_Statement) or (Statement_Kind = A_Case_Statement)) then
                     Nodes.Set_Info(Ei_Node_Ptr.all, S2u("end if/case"));
                  else
                     Nodes.Set_Info(Ei_Node_Ptr.all, S2u("end select"));
                  end if;
                  -- Put the "EndIf" node onto the Stack
                  Add_Node(Ei_Node_Ptr);
                  -- Put the "If" Node onto the Stack again
                  Add_Node(New_Node_Ptr);

               when A_Loop_Statement | A_While_Loop_Statement | A_For_Loop_Statement =>

                  -- Info: ** Loop-Class Nodes (PRE)
                  --
                  --       We replace the general Purpose Node with a Loop one; Because the
                  --       Statement Processing always replaces the top Node on the Node Stack
                  --       AND because we cannot afford to lose Track of the Loop Head (we want
                  --       to be able to link the last Node of the Loop back to the first one),
                  --       we must put this Node twice on the Stack;
                  --
                  --       The Stack Top will become "Loop-Loop". The last Loop will be replaced
                  --       by the Loop Nodes. At the End of the Loop we will link the last Node
                  --       back to the Loop Head;

                  -- Create a new Loop Node
                  if Statement_Kind = A_Loop_Statement then
                     Loop_Node_Ptr := new Nodes.Infinite_Loop_Node_Object;
                  else
                     Loop_Node_Ptr := new Nodes.Finite_Loop_Node_Object;
                  end if;
                  Nodes.Init(Loop_Node_Ptr.all, Flow_Ptr, Element);
                  Nodes.Set_Info(Loop_Node_Ptr.all, S2u(First_Line(Element)));
                  Last_Node_Ptr := Replace_Last_Node_And_Relink(Loop_Node_Ptr, Prev_Node_Ptr);
                  New_Node_Ptr := Loop_Node_Ptr;

                  -- Duplicate the Node on the Stack
                  Add_Node(New_Node_Ptr);

               when A_Return_Statement | A_Terminate_Alternative_Statement =>

                  -- Info: ** Termination-Class Nodes (PRE)
                  --
                  --       We replace the general Purpose Node with a Return/Terminate one

                  -- Create a new Return/Terminate Node
                  if (Statement_Kind = A_Return_Statement) then
                     Ret_Node_Ptr := new Nodes.Return_Node_Object;
                  elsif (Statement_Kind = A_Terminate_Alternative_Statement) then
                     Ret_Node_Ptr := new Nodes.Terminate_Node_Object;
                  end if;
                  Nodes.Init(Ret_Node_Ptr.all, Flow_Ptr, Element);
                  Nodes.Set_Info(Ret_Node_Ptr.all, S2u(First_Line(Element)));
                  Last_Node_Ptr := Replace_Last_Node_And_Relink(Ret_Node_Ptr, Prev_Node_Ptr);
                  -- Reset the new_node_ptr Pointer
                  New_Node_Ptr := Ret_Node_Ptr;

               when others =>
                  null;
            end case;

            -- Info: ** Parameter Tree Handling (PRE)
            --
            --       At the End of the Statement (PRE) Processing we can set the Param Node
            --       as RHS Tree of the current Statement (provided we still have a Param Node
            --       and an Active Node);

            if (Param_Node_Ptr /= null) and (New_Node_Ptr /= null) then
               Nodes.Set_RHS_Tree(New_Node_Ptr.all, Get_Last_Param_Node);
            end if;

         when An_Exception_Handler =>

            -- E X C E P T I O N S

            -- Info: ** Exception Handlers (PRE)
            --
            --       An Exception Handler beginns here; We will only know the Name of the
            --       Exception being caught here when the Child containing it has been processed;
            --       There is nothing we can do about it here;

            null;

         when others =>
            null;
      end case;

   end Handle_Pre;

   -- Post OP
   procedure Handle_Post
     (Element :        Asis.Element;
      Control : in out Asis.Traverse_Control;
      State   : in out Actuals_For_Traversing.Traversal_State) is
      --
      Element_Kind:     Asis.Element_Kinds;
      Clause_Kind:      Asis.Clause_Kinds;
      Association_Kind: Asis.Association_Kinds;
      Definition_Kind:  Asis.Definition_Kinds;
      Declaration_Kind: Asis.Declaration_Kinds;
      Expression_Kind:  Asis.Expression_Kinds;
      Statement_Kind:   Asis.Statement_Kinds;
      --
      Flow_Ptr:         Flow_Types.Flow_Class_Ptr                       := null;
      Pkg_Ptr, New_Pkg_Ptr:
        Pkgs.Pkg_Class_Ptr                                              := null;
      Cfg_Ptr:          CFGs.CFG_Class_Ptr                              := null;
      --
      If_Node_Ptr, Ei_Node_Ptr, Loop_Node_Ptr, Node_Ptr, Prev_Node_Ptr,
        Call_Node_Ptr, Jump_Node_Ptr, New_Node_Ptr:
        Nodes.Node_Class_Ptr                                            := null;
      --
      Full_Name:         Unbounded_String                               := S2u("");
   begin
      -- Print the Implementation Status of the Current Element Type
      Impl_Msg.Display_Message(" * [POST] ",Element);

      -- Remove the Child Element from the Stack
      Close_Last_Element;

      -- Identify the current Structures
      Get_Last_Flow_Info(Flow_Ptr, Pkg_Ptr, Cfg_Ptr);

      -- Info: ** Flow Structure Closing
      --
      --       Factum: Certain ASIS Types generate Packages and other generate CFGs;
      --       Thus, we could close the created Structures when PostOp-ing said Types;
      --       However: certain CFGs may be induced by Types that do not usually imply
      --       a new CFG; Such Cases occur (e.g) in Initialization CFGs where, suddenly,
      --       Statements are found inside a Package;
      --
      --       Since each Flow Object placed on the Flow Stack also contains the Element
      --       that induced that Structure, we will make a general Check to see if the
      --       Element we are just PostOp-ing also induced the current Structure;
      --       Should this be the Case, we will close the current Structure as well;

      -- Check if the current Element generated this Flow Structure
      if (Asis.Elements.Is_Equal(Get_Last_Flow_Elem, Element)) then
         -- Yes, the current Element is responsible for the current Structure
         Put_Line (Debug, "[Transformation] Closing: " & Flow_Types.Get_Name(Flow_Ptr.all));
         -- Close Structure
         Close_Last_Flow;
         -- Update Variable Information
         Get_Last_Flow_Info(Flow_Ptr, Pkg_Ptr, Cfg_Ptr);

         -- Info: ** Init CFG Closing
         --
         --       We must check a second Time for the Generator Element because we might have
         --       just closed the Init CFG for the Current Package and we may still have to close
         --       the Package itself;

         if (Asis.Elements.Is_Equal(Get_Last_Flow_Elem, Element)) then
            Put_Line (Debug, "[Transformation] Closing Structure: " &
                      Flow_Types.Get_Name(Flow_Ptr.all));
            -- Close the Structure
            Close_Last_Flow;
            -- Update the Variables
            Get_Last_Flow_Info(Flow_Ptr, Pkg_Ptr, Cfg_Ptr);
         end if;
      end if;

      -- Check the Element Kind
      Element_Kind := Asis.Elements.Element_Kind (Element);
      case Element_Kind is

         when A_Clause =>

            -- C L A U S E S

            Clause_Kind := Asis.Elements.Clause_Kind (Element);
            case Clause_Kind is

               when A_With_Clause | A_Use_Package_Clause | A_Use_Type_Clause =>

                  -- Info: ** Entity Visibility & Inclusion Clauses (POST)
                  --
                  --       A Visibility Clause has just finished; There isn't much that we can do
                  --       about it at this Point; When the next Flow Structure is being created,
                  --       the Clause List will automatically be appended to it;

                  null;

               when others =>
                  null;
            end case;

         when An_Association =>

            -- A S S O C I A T I O N S

            Association_Kind := Asis.Elements.Association_Kind (Element);
            case Association_Kind is

               when A_Parameter_Association | A_Discriminant_Association |
                 An_Array_Component_Association | A_Record_Component_Association =>

                  -- Info: ** Parameter (Class) Association (POST)
                  --
                  --       A PA has just finished and thus we no longer need the Parameter Node
                  --       created for this Association in the PreOp Phase, in the Parameter
                  --       Tree (the current Stack Top);

                  -- Close the Parameter Node
                  Close_Last_Param;
                  -- Save the Names found on the Way and clear the List
                  UStr_List.Clear(Trans_Data.Names);

               when A_Generic_Association =>

                  -- Info: ** Generic Associations (POST)
                  --
                  --       The GA Records are being appended in the PreOp to their Parent
                  --       Structure and since they have no Stack Handling either, there isn't
                  --       much we can do with them here;

                  -- Make Room for new Names
                  UStr_List.Clear(Trans_Data.Names);

               when others =>
                  null;
            end case;

         when A_Definition =>

            -- D E F I N I T I O N S

            Definition_Kind := Asis.Elements.Definition_Kind (Element);
            case Definition_Kind is

               when A_Type_Definition | A_Formal_Type_Definition | A_Component_Definition |
                 A_Private_Type_Definition | A_Tagged_Private_Type_Definition |
                 A_Private_Extension_Definition =>

                  -- Info: ** (Formal) Type Definitons (POST)
                  --
                  --       A TypeDef has just finished; We must close the Parameter Tree we opened
                  --       in the PreOp Phase;

                  Close_Last_Param;

               when A_Subtype_Indication =>

                  -- Info: ** Subtype Indications (POST)
                  --
                  --       No special Handling needed here;

                  null;

               when A_Record_Definition =>

                  -- Info: ** Record Definitions (POST)
                  --
                  --       The Definition of a Record has just finished; Since we do not create
                  --       any special Container for the Component Declarations (each Component
                  --       will be saved into the Components Field of the new Type) we do not
                  --       need to perform any special Action here;

                  null;

               when A_Discrete_Subtype_Definition =>

                  -- Info: ** Discrete Subtype Definitions (POST)
                  --
                  --       No special Handling needed here;

                  null;

               when A_Task_Definition =>

                  -- Info: ** Task Definitions (POST)
                  --
                  --       No special Handling needed here;

                  null;

               when others =>
                  null;
            end case;

         when A_Declaration =>

            -- D E C L A R A T I O N S

            -- Info: ** Declarations
            --
            --       Declarations usually occur only in the Declarative Section of a Flow Object
            --       and since Nodes only exist in the Imperative Part it makes little Sense to
            --       try and save the Identifiers found so far, into the previous Node;
            --
            --       However, like any Rule, this one has its Exceptions too; e.g. The Loop
            --       Parameter Specification - the Structure containing the Parameter of a Loop
            --       and its Interval - IS a Declaration, and may contain Variables that would
            --       otherwise get lost if we did not save them;
            --
            --       This saving Point for Identifiers will cause a lot of Warnings, but will
            --       also insure that Identifier Losses are minimal;

            -- Save found Identifiers
            Save_Identifiers(Trans_Data.Names);

            Declaration_Kind := Asis.Elements.Declaration_Kind (Element);
            case Declaration_Kind is

               when An_Ordinary_Type_Declaration | A_Subtype_Declaration |
                 A_Private_Type_Declaration | A_Private_Extension_Declaration =>

                  -- Info: ** (Sub-)Type Declarations (POST)
                  --
                  --       When a Type Declaration finishes we must save the new Type into
                  --       the current Flow Object and reset the Data Pointer;

                  Flow_Types.Add_Type(Flow_Ptr.all, Trans_Data.Type_Decl);
                  Trans_Data.Type_Decl := null;

               when A_Package_Declaration | A_Package_Body_Declaration |
                 A_Protected_Type_Declaration | A_Single_Protected_Declaration |
                 A_Protected_Body_Declaration =>

                  -- Info: ** Flow Package Structure Declarations (POST)
                  --
                  --       The Packages these ASIS Types have induced have already been closed
                  --       in the Beginning of the PostOp Method

                  null;

               when A_Variable_Declaration | A_Component_Declaration |
                 A_Parameter_Specification | An_Entry_Index_Specification |
                 A_Discriminant_Specification | A_Loop_Parameter_Specification |
                 A_Constant_Declaration | A_Deferred_Constant_Declaration =>

                  -- Info: ** Variable-Class Declarations (POST)
                  --
                  --       When PostOp-ing a such a Declaration, we must transfer the gathered
                  --       Information to the current Flow Object (*);
                  --       We then close the Declaration Structure created in the PreOp;
                  --       The Parameter Tree created for the possible Initialization of the
                  --       Variable must also be closed (Note: for all BUT the Loop ParSpec!)
                  --
                  --       (*) There are, however, Exceptions from the Flow Object Rule:
                  --
                  --       1. Component Declarations also occur in Record Definitions; Since no
                  --          other Declaration Kind can be nested into a Record Definition, we
                  --          have to save the Component into the Record instead of the Flow, if
                  --          any Record Structure is present;
                  --
                  --          Note: Anonymous Record Definitions are not allowed; This means that
                  --          when we are dealing with a RecDef, we are also dealing with a Type
                  --          Declaration which in turn means that we currently have an open Type
                  --          Description Structure;
                  --
                  --          In Conclusion: if there is an open Type Description Structure, we
                  --          must append the current (Component) Declaration to it, instead of
                  --          appending it to the current Flow Object;
                  --
                  --       2. Discriminant Specifications occur not only in Task Declarations but
                  --          also in Type (!) Declarations; Same as above, if a Type Description
                  --          Structure is present, is has Precedence over the Flow Object;
                  --
                  --       We will also dispose of the Pointer to last Parameter Tree here since
                  --       the Trees we have just been working with do not need referencing from
                  --       anywhere else;

                  -- Transfer Data
                  if ((not Trans_Data.Ignore_Mode) and
                      (not ((Declaration_Kind = A_Parameter_Specification) and
                            (Trans_Data.Type_Decl /= null)))) then

                     if ((Declaration_Kind = A_Parameter_Specification) or
                         (Declaration_Kind = An_Entry_Index_Specification) or
                         (Declaration_Kind = A_Discriminant_Specification)) then
                        if ((Trans_Data.Type_Decl /= null) and then
                            (Declaration_Kind = A_Discriminant_Specification)) then
                           -- This is a Discriminat of the new Type
                           Flow_Types.Declaration_Ptr_List.Append(Trans_Data.Type_Decl.Discriminants,
                                                                  Trans_Data.Decl);
                        elsif (Flow_Ptr.all in CFGs.CFG_Object'Class) then
                           -- These were Subprogram Parameters, not Variables
                           CFGs.Add_Param(CFGs.CFG_Object(Flow_Ptr.all), Trans_Data.Decl);
                        else
                           Put_Line(Error, "[Transformation] [WARN] " &
                                    "Parameter / Entry / Discriminant without Parent Structure!");
                        end if;
                     elsif (Declaration_Kind = A_Loop_Parameter_Specification) then
                        -- This is a Loop Control Variable; We must check if we have a Node open
                        Loop_Node_Ptr := Get_Last_Node;
                        if (Loop_Node_Ptr /= null) then
                           -- This was a Loop Parameter Specification -> Add it to the Node
                           Flow_Types.Add_Decl(Flow_Types.Flow_Object(Loop_Node_Ptr.all),
                                               Trans_Data.Decl);
                        else
                           Put_Line(Error,"[Transformation] [WARN]" &
                                    " Loop Parameter Specification without Father Node!");
                        end if;
                     else
                        -- These were Variables or Components
                        if (Trans_Data.Type_Decl /= null) then
                           -- This was a Record Component -> add it to the Record
                           Flow_Types.Declaration_Ptr_List.Append(Trans_Data.Type_Decl.Components,
                                                                  Trans_Data.Decl);
                        else
                           -- No Type Structure open -> must be either a Variable or a protected
                           -- Object Component -> Add it to the Flow
                           Flow_Types.Add_Decl(Flow_Ptr.all, Trans_Data.Decl);
                        end if;
                     end if;

                     -- Clear Variables
                     Trans_Data.Decl := null;
                     -- Clear the Parameter Tree
                     if (Declaration_Kind /= A_Loop_Parameter_Specification) then
                        Close_All_Params;
                     end if;

                     Trans_Data.Last_PTree := null;

                  end if;

               when A_Procedure_Declaration | A_Function_Declaration |
                 An_Entry_Declaration |
                 A_Procedure_Body_Stub | A_Function_Body_Stub | A_Task_Body_Stub =>

                  -- Info: ** CFG-Class Spec Declarations (POST)
                  --
                  --       These Structures have finished and we can now reset the Ignore_Mode,
                  --       so that Parameter Declarations will be taken into Consideration again;

                  Trans_Data.Ignore_Mode := False;

               when A_Procedure_Body_Declaration | A_Function_Body_Declaration |
                 An_Entry_Body_Declaration |
                 A_Single_Task_Declaration | A_Task_Type_Declaration | A_Task_Body_Declaration =>

                  -- Info: ** CFG-Class Body Declarations (POST)
                  --
                  --       The CFGs induced by these Structures have already been closed in the
                  --       Beginning of this Method;

                  null;

               when A_Generic_Package_Declaration |
                  A_Generic_Procedure_Declaration | A_Generic_Function_Declaration =>

                  -- Info: ** Generic Structures (POST)
                  --
                  --       A Generic Structure has just finished; We needn't do anything since
                  --       the Package / CFG hast already been closed earlier;

                  null;

               when A_Formal_Object_Declaration | A_Formal_Type_Declaration |
                 A_Formal_Procedure_Declaration | A_Formal_Function_Declaration |
                 A_Formal_Package_Declaration | A_Formal_Package_Declaration_With_Box =>

                  -- Info: ** Formal Declarations (POST)
                  --
                  --       We have just finished a Formal Declaration; There is nothing we can
                  --       do about it at this Point; Its Content has been added and the Formal
                  --       Declaration Structure will be appended to the Flow Structure when
                  --       the Generic FS is being initialized;
                  --       We must only reset the Ignore Flag so Declarations will be taken into
                  --       consideration again;
                  --       The Parameter Tree created for the possible Initialization of the
                  --       Variable must also be closed;

                  Trans_Data.Ignore_Mode := False;

                  -- Clear the Parameter Tree
                  Close_All_Params;

               when A_Procedure_Renaming_Declaration | A_Function_Renaming_Declaration |
                 A_Generic_Procedure_Renaming_Declaration | A_Generic_Function_Renaming_Declaration |
                 A_Package_Renaming_Declaration | A_Generic_Package_Renaming_Declaration |
                 An_Object_Renaming_Declaration | An_Exception_Renaming_Declaration  =>

                  -- Info: ** Entity Renaming Declarations (POST)
                  --
                  --       A Renaming has just finished; Usually we save Information in the
                  --       HCHF-Method, but in this Case we need to make an Exception;
                  --       We know only that the _LAST_ Child of a Renaming Declaration is
                  --       Name of the Old Structure; We do not know how many Children there
                  --       are and thusly we cannot save the Name of the old Structure in the
                  --       HCHF-Method, but only after the entire Structure has finished;
                  --       We also no longer need to ignore Parameter Declarations;

                  Trans_Data.Ignore_Mode := False;

                  if not UStr_List.Empty(Trans_Data.Rename.Old_Name) then
                     -- An Old Name already exists, so do nothing!
                     null;
                  elsif (not UStr_List.Empty(Trans_Data.Names)) then
                     -- Save the Old Name of the Object
                     Full_Name := UStr_List.Get_Tail(Trans_Data.Names);
                     Trans_Data.Rename.Old_Name := Tokenize(U2s(Full_Name), ".");
                  else
                     Put_Line(Warning, "[Transformation] [WARN] " &
                              "Cannot retrieve Old Name for Renaming!");
                  end if;

                  Close_Last_Param;

                  -- Save the Renaming..
                  if (Flow_Ptr = null) then
                     -- .. hmm, there is NO Flow Object! Bummer..
                     -- Just like in the Case of the CFGs w/o Pkg we will create a Default one
                     New_Pkg_Ptr := new Pkgs.Def_Object;
                     Pkgs.Init(New_Pkg_Ptr.all, W2u("Default_PKG"), null, Element);
                     Flow_Types.Set_Refs(Flow_Types.Flow_Object(New_Pkg_Ptr.all), Trans_Data.Refs);
                     Put_Line (Debug,"[Transformation] New PKG: " & Pkgs.Get_Name(New_Pkg_Ptr.all));
                     Flow_World.Add_Pkg(World, New_Pkg_Ptr);
                     -- Put the new Package onto the Flow Stack
                     Open_Flow(Flow_Types.Flow_Class_Ptr(New_Pkg_Ptr), Element);
                     -- Update Variables
                     Get_Last_Flow_Info(Flow_Ptr, Pkg_Ptr, Cfg_Ptr);
                  end if;

                  -- .. into the Flow Object
                  Flow_Types.Add_Ren(Flow_Ptr.all, Trans_Data.Rename);
                  -- Clear the Renaming Structure
                  Trans_Data.Rename := null;

               when A_Package_Instantiation |
                 A_Procedure_Instantiation | A_Function_Instantiation =>

                  -- Info: ** Generic Instantiations (POST)
                  --
                  --       An Instantiation has just Finished; We must add it to the current
                  --       Flow Structure; Since Renamings and Instantiations share somewhat
                  --       similar Semantics, we will use the same Renaming Structures to keep
                  --       both of them;

                  -- Save the Renaming into the Flow Object
                  if Flow_Ptr /= null then
                     Flow_Types.Add_Ren(Flow_Ptr.all, Trans_Data.Rename);
                  elsif Declaration_Kind = A_Package_Instantiation then
                     Put_Line(Debug, "[Transformation] A package instantiation has occurred out of context - will create a new package.");
                     New_Pkg_Ptr := new Pkgs.Simple_Spec_Object;
                     Pkgs.Init(New_Pkg_Ptr.all, Trans_Data.Rename.New_Name, null, Element);
                     Flow_World.Add_Pkg(World, New_Pkg_Ptr);
                     Flow_Ptr := Flow_Types.Flow_Object(New_Pkg_Ptr.all)'Access;
                     Flow_Types.Add_Ren(Flow_Ptr.all, Trans_Data.Rename);
                  else
                     Put_Line(Error, "[Transformation] A subprogram instantiation has occurred out of context!");
                     Flow_Types.Free(Trans_Data.Rename);
                  end if;

                  -- Clear the Renaming Structure
                  Trans_Data.Rename := null;

               when others =>
                  null;
            end case;

            -- Make Room for new Names
            UStr_List.Clear(Trans_Data.Names);

         when An_Expression =>

            -- E X P R E S S I O N S

            Expression_Kind := Asis.Elements.Expression_Kind (Element);
            case Expression_Kind is

               when A_Positional_Array_Aggregate | A_Record_Aggregate =>

                  -- Info: ** Aggregates (POST)
                  --
                  --       We must dispose of the Node created in the PreOp for this Aggregate;

                  Close_Last_Param;

               when An_Attribute_Reference =>

                  -- Info: ** Attribute References (POST)
                  --
                  --       ARs are "Var'Attribute" Type Constructs; Their *last* Child is the
                  --       Name of the Attribute; It is of ASIS Type Expression/Identifier;
                  --       Since Attributes can only be read, and never directly written, we do
                  --       not need them; We will thusly remove the last saved Identifier from
                  --       the List;

                  UStr_List.Remove_Last(Trans_Data.Names);

               when An_Allocation_From_Subtype =>

                  -- Info: ** Allocation From Subtype (POST)
                  --
                  --       An AFS has finished and we must close the Parameter Allocation Node
                  --       we have placed earlier onto the Parameter Stack;

                  -- Close the Parameter Node
                  Close_Last_Param;
                  -- Save the Names found on the Way and clear the List
                  Save_Identifiers(Trans_Data.Names);
                  UStr_List.Clear(Trans_Data.Names);

               when A_Type_Conversion =>

                  -- Info: ** Type Conversions (POST)
                  --
                  --       The Type Conversion is the typical Type Casting; There is no
                  --       Post-Operation we can do here; The Work is done in HCHF;

                  null;

               when An_Identifier =>

                  -- Info: ** Identifirs (POST)
                  --
                  --       We must save each found Identifier into the Names List for processing;

                  UStr_List.Append(Trans_Data.Names, S2u(First_Line(Element)));

               when An_Operator_Symbol =>

                  -- Info: ** Operator Symbols (POST)
                  --
                  --       We must save each found Operator into the Names List for processing;

                  UStr_List.Append(Trans_Data.Names, S2u(First_Line(Element)));

               when A_Function_Call =>

                  -- Info: ** Function Calls (POST)
                  --
                  --       A Function Call has just finished; We must close the current Object
                  --       on the Parameter Stack - the one we created for this Function in the
                  --       PreOp-Phase; Also, we will add the Names found in this Subtree
                  --       (Variables, ..) to the current Node (provided this Function Call was
                  --       done in a Statement and not in a Declaration) and then clear the List;

                  -- Close Param Node
                  Close_Last_Param;

                  -- Save the found Names into the current Node and clear the List
                  Save_Identifiers(Trans_Data.Names);
                  UStr_List.Clear(Trans_Data.Names);

               when A_Selected_Component =>

                  -- Info: ** Selected Components (POST)
                  --
                  --       Selected Components are reccursive Structures in the AST and the
                  --       Finishing of one Structures bears no Information about the whole
                  --       Subtree, so there is nothing we can do about them here;

                  null;

               when An_Indexed_Component | A_Slice =>

                  -- Info: ** Indexed (-Class) Components (POST)
                  --
                  --       Indexed Components may also be nested, complex Structures and their
                  --       Usage is Context Sensitive, so when one finishes, there is nothing we
                  --       can do, as long as we have no Information regarding the Structures
                  --       surrounding it;
                  --
                  --       The leading Name has already been pushed into the current Parameter
                  --       Node, and if any Functions have been used, they will also have already
                  --       been saved; We now clear the Identifiers List to ensure that no
                  --       incorrect Names are being created;
                  --
                  --       We also close the Parameter Node created for the Component;

                  Close_Last_Param;

                  -- Save the found Names into the current Node and clear the List
                  Save_Identifiers(Trans_Data.Names);
                  UStr_List.Clear(Trans_Data.Names);

               when others =>
                  null;
            end case;

         when A_Path =>

            -- P A T H S

            -- Info: ** Paths (POST)
            --
            --       A Path has just finished; We could use this Trigger to reconstruct the
            --       Structure of the Node Stack here, but we will do it in the HCHF Method,
            --       when triggered by the Termination of a Child of a Branch Class Statement;
            --
            --       Note that the Paths ARE the Children of the Branch Class Statements, so we
            --       are only postponing the Transformation of the Stack by one (1) Procedure
            --       Call;

            -- Save Identifiers
            Save_Identifiers(Trans_Data.Names);

         when A_Statement =>

            -- S T A T E M E N T S

            -- ** GENERAL STATEMENT (POST) HANDLING

            -- Info: ** Statements (POST)
            --
            --       When a Statement finishes, all the Information about it will have already
            --       been saved into the associateed CFG Node and there isn't much left to do
            --       about said Statement;

            -- Info: ** Identifier Policy when (POST) processing Statements
            --
            --       Since we have been gathering the Names found in this Statement in Name List,
            --       the PostOp is the last Chance we have to save them into the Node itself
            --       before the List gets cleared for the next Statement;

            -- Save Identifiers
            Save_Identifiers(Trans_Data.Names);

            -- Info: ** Parameter Tree (POST) Handling
            --
            --       We currently have a Parameter Tree open on the Stack for this Statement;
            --       Since the Statement is now over, we no longer need the Tree;

            Close_All_Params;

            -- ** SPECIAL STATEMENT (POST) HANDLING

            -- Info: ** Statements (POST)
            --
            --       The only Statements that need special PostOp Handling are the ones that
            --       used an exotic Stack Order; We now need to restore the Stack Top Order
            --       and add finishing Touches to the Nodes themselves, if necessary;

            Statement_Kind := Asis.Elements.Statement_Kind (Element);
            case Statement_Kind is

               when An_Assignment_Statement =>

                  -- Info: ** Assignment Statements (POST)
                  --
                  --       No Special Handling needed;

                  null;

               when A_Goto_Statement | An_Exit_Statement =>

                  -- Info: ** Jump-Class Statements (POST)
                  --
                  --       No Special Handling needed;

                  null;

               when A_Procedure_Call_Statement | A_Block_Statement | A_Raise_Statement |
                 An_Entry_Call_Statement | A_Requeue_Statement | A_Requeue_Statement_With_Abort |
                 An_Abort_Statement | An_Accept_Statement =>

                  -- Info: ** Call-Class Statements Statements (POST)
                  --
                  --       No Special Handling needed;

                  null;

               when An_If_Statement | A_Case_Statement | A_Selective_Accept_Statement |
                 A_Conditional_Entry_Call_Statement | A_Timed_Entry_Call_Statement |
                 An_Asynchronous_Select_Statement =>

                  -- Info: ** Branch-Class Nodes (POST)
                  --
                  --       The BC Nodes use an "If-EndIf-If" Stack Top Order; According to
                  --       the Control Flow Semantics, the last Node in the CFG, after the
                  --       Processing of a BC Node must be the "EndIf" Node, marking the End
                  --       of the Branching; We thusly have to remove all three Top Nodes from
                  --       the Node Stack and place the "EndIf" Node on Top;

                  -- Rearrange Stack
                  If_Node_Ptr := Remove_Last_Node;
                  Ei_Node_Ptr := Remove_Last_Node;
                  If_Node_Ptr := Remove_Last_Node;
                  Add_Node(Ei_Node_Ptr);

                  -- Info: ** Forced Else Paths
                  --
                  --       An "If" Statement doesn't HAVE to have an Else Branch; Without an ASIS
                  --       Else Path Node to mark the Beginning of such a Branch we would not add
                  --       one; In such Cases we would thusly not have an alternative Path should
                  --       the If Check fail;
                  --
                  --       To prevent constructing an erroneous CFG we will force adding an empty
                  --       Else Path should ASIS not provide any;
                  --
                  --       During the Transformation we mark the "If" Statements that have an
                  --       Else Path with a Value of "2"; Should thusly an "If" Statement have
                  --       no "2" Mark when it finishes, it would consequently mean it has no
                  --       Else Branch, and we must add one ourselves; This means linking the
                  --       "If" node directly to the "EndIf" one;

                  -- Only "If" Statements need an Else - Case Statements don't
                  if (Statement_Kind = An_If_Statement) then
                     if (Nodes.Get_Mark(If_Node_Ptr.all) /= 2) then
                        -- Link the "If" Node directly to the "EndIf" one
                        Nodes.Add_Succ (If_Node_Ptr.all, Flow_Types.Flow_Class_Ptr(Ei_Node_Ptr));
                        Nodes.Add_Pred (Ei_Node_Ptr.all, Flow_Types.Flow_Class_Ptr(If_Node_Ptr));
                     end if;
                  end if;

               when A_Loop_Statement | A_While_Loop_Statement | A_For_Loop_Statement =>

                  -- Info: ** Loop-Class Nodes (POST)
                  --
                  --       The LC Statements use a "LoopHead"-"LoopNode" Stack Top Order;
                  --       According to the Control Flow Semantics, the last Node in the CFG,
                  --       after the Processing of a LC Node must be the "LoopHead" Node;
                  --       We need to remove the last LoopNode from the Stack AND we also need
                  --       to link it back to the Loop Head - after the Execution of the last
                  --       Node in the Loop, the Control passes back to the Loop Head;
                  --
                  --       To make sure that the End-Node of a CFG is not part of a Loop, we will
                  --       add a pseudo-node after each Loop;

                  -- Rearrange Stack
                  Node_Ptr := Remove_Last_Node;
                  -- Link the last Node backwards to the Loop Head
                  Loop_Node_Ptr := Get_Last_Node;
                  Nodes.Add_Succ(Node_Ptr.all, Flow_Types.Flow_Class_Ptr(Loop_Node_Ptr));
                  Nodes.Add_Pred (Loop_Node_Ptr.all, Flow_Types.Flow_Class_Ptr(Node_Ptr));

                  -- Create the pseudo Node, set the Node Information and save it
                  New_Node_Ptr := new Nodes.Node_Object;
                  Nodes.Init(New_Node_Ptr.all, Flow_Ptr, Element);
                  Nodes.Set_Info(New_Node_Ptr.all, S2u("Loop End"));
                  -- Replace the Top of the Node Stack with the New Node
                  Prev_Node_Ptr := Replace_Last_Node(New_Node_Ptr);

               when A_Return_Statement | A_Terminate_Alternative_Statement =>

                  -- Info: ** Termination-Class Nodes (POST)
                  --
                  --       No Special Handling needed;

                  null;

               when others =>
                  null;
            end case;

            -- Make Room for more Identifiers
            UStr_List.Clear(Trans_Data.Names);

         when An_Exception_Handler =>

            -- E X C E P T I O N S

            -- Info: ** Exception Handlers (POST)
            --
            --       An Exception Handler has just finished; Its CFG will be closed automatically
            --       so there isn't much we can do about it here;

            null;

         when others =>
            null;
      end case;

      -- Info: ** Statement Termination Notification
      --
      --       Since the Purpose of an AST Node only becomes clear in a Context, we need a
      --       Trigger for each AST Node when one of its Children has finished; For that Purpose,
      --       each time a Node finishes (thusly in the PostOp) we will try to notify its Father;

      -- Peek who the Father of the current Node is
      if (not Asis.Elements.Is_Equal(Get_Last_Element, Nil_Element)) then
         -- Increase the Number of the Children that have finished by 1
         Set_Child_Finished;
         -- Notify the Father
         Handle_Child_Finish(Get_Last_Element, Element, Get_Last_Child_Nr, State);
      end if;
   end Handle_Post;

   -- Child Post OP
   procedure Handle_Child_Finish
     (Parent : Asis.Element;
      Child  : Asis.Element;
      Count  : in Integer;
      State  : in out Actuals_For_Traversing.Traversal_State) is
      --
      Element_Kind, Child_Element_Kind: Asis.Element_Kinds;
      Association_Kind:                 Asis.Association_Kinds;
      Clause_Kind:                      Asis.Clause_Kinds;
      Definition_Kind:                  Asis.Definition_Kinds;
      Declaration_Kind:                 Asis.Declaration_Kinds;
      Expression_Kind:                  Asis.Expression_Kinds;
      Statement_Kind:                   Asis.Statement_Kinds;
      --
      Flow_Ptr:                         Flow_Types.Flow_Class_Ptr := null;
      Cfg_Ptr, New_Cfg_Ptr:             CFGs.CFG_Class_Ptr        := null;
      Pkg_Ptr, New_Pkg_Ptr:             Pkgs.Pkg_Class_Ptr        := null;
      --
      Node_Ptr, If_Node_Ptr, Ei_Node_Ptr, Call_Node_Ptr,
        Jump_Node_Ptr, Param_Node_Ptr, Prev_Node_Ptr:
        Nodes.Node_Class_Ptr                                      := null;
      --
      Component_Name, Full_Name:        Unbounded_String          := S2u("");
      Name_List:                        UStr_List.Object;
      Local_Alias:                      Alias_Class_Ptr           := null;
      Inst_Alias:                       Instance_Alias_Class_Ptr  := null;
      Form_Decl:                        Formal_Decl_Class_Ptr     := null;
      Ref_Ptr:                          Flow_Types.Ref_Class_Ptr  := null;
      Type_Ptr:                         Flow_Types.Type_Descr_Ptr := null;
   begin
      -- Print the Implementation Status of the Current Element Type
      Impl_Msg.Display_Message(" * [HCHF] ",Parent);

      -- Identify the current Structures
      Get_Last_Flow_Info(Flow_Ptr, Pkg_Ptr, Cfg_Ptr);

      -- Check the Element Kind
      Element_Kind := Asis.Elements.Element_Kind (Parent);
      case Element_Kind is

         when A_Clause =>

            -- C L A U S E S

            Clause_Kind := Asis.Elements.Clause_Kind (Parent);
            case Clause_Kind is

               when A_With_Clause | A_Use_Package_Clause | A_Use_Type_Clause =>

                  -- Info: ** Entity Visibility & Inclusion Clauses (HCHF)
                  --
                  --       The Visibility Clauses have only one Child, and the Child has the Type
                  --       Expression (Identifier or Selected Component); We will save this Child
                  --       into the last Reference Structure available;

                  Ref_Ptr := Flow_Types.Ref_Class_Ptr_List.Get_Tail(Trans_Data.Refs);
                  if (Ref_Ptr = null) or (Count /= 1) or (UStr_List.Empty(Trans_Data.Names)) then
                     Put_Line (Error, "[Transformation] [ERR] " &
                               "Clause Inconsistency: Null Pointer or wrong Structure!");
                  else
                     Full_Name := UStr_List.Get_Head(Trans_Data.Names);
                     Ref_Ptr.Name := Tokenize(U2s(Full_Name), ".");
                  end if;

               when others =>
                  null;
            end case;

         when An_Association =>

            -- A S S O C I A T I O N S

            Association_Kind := Asis.Elements.Association_Kind (Parent);
            case Association_Kind is

               when A_Parameter_Association | A_Discriminant_Association |
                 An_Array_Component_Association | A_Record_Component_Association =>

                  -- Info: ** Parameter (Class) Associations (HCHF)
                  --
                  --       A PA Subtree has one or two Children, both Expressions; The last one
                  --       is always the Expression to be used as Value for the Parameter; The
                  --       first Child (if there are two) is always the Name of the Parameter;
                  --
                  --       The standard Action here is to save the found Identifiers as VALUE;
                  --       Before doing so, however, we must check if this is the SECOND Child
                  --       of the PA; If this is the Case, it must mean that, when processing
                  --       the FIRST Child (i.e. the Name of the Parameter), we mistakingly saved
                  --       it (per Default) into the Value instead of the Name; For that we will
                  --       now move the Value into the Name and only then proceed with the
                  --       standard Action;
                  --
                  --       There are several ASIS Types that may occur as Value; Immidiates (like
                  --       Constants) are being ignored by the Transformation Logic, because they
                  --       are no Identifiers, and thus do not appear in the Name List;
                  --       Variable-like Constructs will be saved in the Var_Name Field of the
                  --       Parameter Node; Function Calls do not get saved in the Var_Name Field,
                  --       because before calling HCHF on the Associtation, a PostOp on the
                  --       Function itself is being called first and the Name List will be
                  --       cleared; This however IS the intedended Behavior because a Function
                  --       Call will appear as a Successor of the actual Parameter Node in the
                  --       current Parameter Tree, anyway;
                  --
                  --       Note that the Identifier we are using for the Value of the Parameter
                  --       (i.e. as Variable Name) could also be a Type Name if we are dealing
                  --       with a Conversion; A more precise Check could improve the Result;

                  if Count = 2 then
                     -- Mutate the Name of the Parameter from "Value" to "Name"
                     Full_Name := S2u(Nodes.Get_Var_Name(Nodes.Param_Node_Object(Get_Last_Param_Node.all)));
                     Nodes.Set_Name(Nodes.Param_Node_Object(Get_Last_Param_Node.all), Full_Name);
                     Nodes.Clear_Var(Nodes.Param_Node_Object(Get_Last_Param_Node.all));
                  end if;

                  -- Default Action: Save the found Identifiers as Parameter Value
                  if (not UStr_List.Empty(Trans_Data.Names)) and
                    (Nodes.Get_Var_Name(Nodes.Param_Node_Object(Get_Last_Param_Node.all)) = "") then
                     -- We have Names and no Variable Name has been set yet
                     Nodes.Add_Var_Item(Nodes.Param_Node_Object(Get_Last_Param_Node.all),
                                        UStr_List.Get_Head(Trans_Data.Names));
                     -- Make Room for new Names
                     UStr_List.Clear(Trans_Data.Names);
                  end if;

               when A_Generic_Association =>

                  -- Info: Generic Associations (HCHF)
                  --
                  --       A GA Node has one or two Children, both Expressions; The last one
                  --       is always the Surrogate Type to be used instead of the Generic one;
                  --       The first Child (if there are two) is always the Name of the Generic
                  --       Type;
                  --
                  --       Generic Associations occur in Generic Instantiations AND in Formal
                  --       Declarations; The Data for a Generic Instantiation is being held in
                  --       the Renaming Structure of the Transformation Data Container; The
                  --       Formal Declarations are being held (since there may be more at the
                  --       same Time (before a Generic Structure claims them)) in the Form_Decls
                  --       List of the same TDC; Since we are using the same Program Code for
                  --       both we will have to identify the Type of Generic Association we are
                  --       dealing with, first;
                  --
                  --       Similar to the Parameter Associations, the Default Action is to try
                  --       and save the found Identifiers as Surrogate Type; Before doing that
                  --       though, we must Check if we are processing the SECOND Child; If this
                  --       is the Case, it then means we have mistakingly saved the first Child
                  --       as Surrogate Type instead of saving it as Generic Type; We will move
                  --       it to the Generic Type and then proceed with the Default Action;
                  --
                  --       Note: We are using Alias Structures, with the New_Name holding the
                  --       Generic and the Old_Name holding the Surrogate Type;

                  -- Identify the Association to use
                  if ((Trans_Data.Rename /= null) and then
                      (Trans_Data.Rename.all in Flow_Types.Instance_Alias'Class)) then
                     -- The current Renaming Structure holds an Instantiation Record
                     Inst_Alias := Instance_Alias(Trans_Data.Rename.all)'Access;
                     Local_Alias := Flow_Types.Alias_Class_Ptr_List.Get_Tail(Inst_Alias.Assocs);
                  elsif (Flow_Types.Formal_Decl_Class_Ptr_List.Count(Trans_Data.Form_Decls) > 0) then
                     -- We have Formal Declarations -> Pick the last one
                     Form_Decl := Flow_Types.Formal_Decl_Class_Ptr_List.Get_Tail(Trans_Data.Form_Decls);
                     Local_Alias := Flow_Types.Alias_Class_Ptr_List.Get_Tail(Form_Decl.Assocs);
                  else
                     Local_Alias := null;
                     Put_Line(Warning,"[Transformation] [WARN] " &
                              "Generic Association without Instantiation Alias!");
                  end if;

                  if (Local_Alias /= null) then
                     if (Count = 2) then
                        -- Mutate erroneous Surrogate to Generic Type
                        Local_Alias.New_Name := UStr_List.Get_Head(Local_Alias.Old_Name);
                        -- Clear the Old Name
                        UStr_List.Clear(Local_Alias.Old_Name);
                     end if;
                     if (not UStr_List.Empty(Trans_Data.Names)) then
                        -- Default Action: Save the Surrogate Type
                        Full_Name := UStr_List.Get_Head(Trans_Data.Names);
                        Local_Alias.Old_Name := Tokenize(U2s(Full_Name), ".");
                        -- Make Room for new Names
                        UStr_List.Clear(Trans_Data.Names);
                     end if;
                  end if;

               when others =>
                  null;
            end case;


         when A_Definition =>

            -- D E F I N I T I O N S

            -- Info: ** Definition Kinds (HCHF)
            --
            --       We will mainly use the Definition Kinds to process Type Definitions;
            --       That is why we will try to find the current Type Description Structure
            --       here;
            --
            --       Since a Type Description may contain Records, and since Records may
            --       contain Component Declarations that, in turn, have Types themselves,
            --       we will first attempt to use the Fields of an alleged Variable/Component
            --       Declaration here and only when this fails do we look for a Type Description;
            --
            --       !!!WARNING!!! The Order of the following Checks is *ESSENTIAL*!
            --        - If we are dealing with a Variable Declaration (or something similar),
            --          it makes no diference what we look at first, because there will not be
            --          any Type Description Structure available in the Transformation Data Cont.
            --        - If we are dealing with Type Definitions (or something similar) that
            --          however *DO NOT* contain Records, it also makes no Difference what we
            --          we look at first, because there will be no Declaration Structures open.
            --        - *IF*, however, we are dealing with a Type Definition (et al.) that *DOES*
            --          contain Record Definitions (that, in turn, contain Component Declarations)
            --          we must pay much Attention to how we handle the Case!
            --          Component Declarations spawn new Declaration Structures, that have their
            --          *OWN* Type Description Structures; When processing Type Related
            --          Definitions, we must make sure that we append them to the Components'
            --          Type Description and not to the main Type's Description; This is always
            --          the correct Handling in such Cases, because Components can be nested into
            --          Type Definitions but never vice-versa;

            -- Try to identify the current open Structure
            if (Trans_Data.Decl /= null) then
               Type_Ptr := Trans_Data.Decl.VType'Access;
            elsif (not Formal_Decl_Class_Ptr_List.Empty(Trans_Data.Form_Decls)) then
               Type_Ptr := Formal_Decl_Class_Ptr_List.Get_Tail(Trans_Data.Form_Decls).VType'Access;
            elsif (Trans_Data.Type_Decl /= null) then
               -- This must be the *LAST* Check!
               Type_Ptr := Trans_Data.Type_Decl;
            else
               Type_Ptr := null;
            end if;

            Definition_Kind := Asis.Elements.Definition_Kind (Parent);
            case Definition_Kind is

               when A_Type_Definition | A_Formal_Type_Definition |
                 A_Private_Type_Definition | A_Tagged_Private_Type_Definition |
                 A_Private_Extension_Definition =>

                  -- Info: ** (Formal) Type Definitons (HCHF)
                  --
                  --       Note: The Considerations below apply to both Standard and Formal Types;
                  --
                  --       Type Definitions (if present) are always the *Top* of the Hierarchy;
                  --       They are not present if this is a simple Variable Declaration without
                  --       any "special Features";
                  --
                  --       Possible Children are:
                  --        1. Subtype Indications
                  --           This means the new Type is a Subtype of the given one;
                  --        2. Component Definitions
                  --           This means the new Type is made up of Items of the given one;
                  --        3. Record Definitions
                  --           This means we are dealing with a new Record-based Type here;
                  --        4. Something else
                  --
                  --       Each Case handles itself (it writes the Data into the correct Slot
                  --       of the Transformation Data Container), so there is nothing to be done
                  --       here, except to save the Identifiers found on the Way;
                  --

                  -- Default Action: Try to add the Identifiers found on the Way
                  Save_Identifiers (Trans_Data.Names);
                  UStr_List.Clear(Trans_Data.Names);

               when A_Discrete_Subtype_Definition =>

                  -- Info: ** Discrete Subtype Definition (HCHF)
                  --
                  --       For the defined Range, we only have to save the found Identifiers;
                  --       If there have been any Functions on the Way, they will have already
                  --       been saved into the correct Parameter Tree;

                  -- Attempt to save the Identifiers
                  Save_Identifiers (Trans_Data.Names);
                  UStr_List.Clear(Trans_Data.Names);

               when A_Subtype_Indication =>

                  -- Info: ** Subtype (Class) Indications (HCHF)
                  --
                  --       A Subtype Indication means that the Type currently being described is
                  --       a Subtype of a given another; The Child containing the Name of the
                  --       Supertype is an Expression/Identifier or */Selected Component;
                  --
                  --       The Default Action is to write this Type into the Supertype Variable
                  --       of whatever Structure is currently open: [Formal] Declaration or
                  --       Type Description;
                  --
                  --       Since there may be Identifiers around we also try to save them;
                  --
                  --       !NOTE! Subtype Indications may occur in four different Scenarios:
                  --               1. Marking an Inheritance Relationship, in which Case we will
                  --                  have a valid Type Description Structure to write the Type
                  --                  Name to;
                  --               2. Marking an Allocation From Subtype IN a Type Definition, in
                  --                  which Case we will have a TDS, BUT we are not be allowed
                  --                  to overwrite the Super Type;
                  --               3. Marking an Allocation from Subtype in a Variable-like
                  --                  Declaration, in which Case we will also have a TDS that we
                  --                  are not allowd to touch (just like 2.);
                  --               4. Marking an Allocation from Subtype in a Statement, in which
                  --                  Case we do not have a TDS at all;
                  --
                  --       Since an AFS handles similarly to a Function Call we will place the
                  --       Name of the allocated Type onto the Parameter Stack; Thus, in order to
                  --       catch the Cases (2.-4.) we will have to check the Parameter Stack for
                  --       a Parameter Allocation Node (that we will have put on the Stack in the
                  --       PreOp) first and should we find one, then this Structure has
                  --       PRECEDENCE over the TDS!

                  if ((Asis.Elements.Element_Kind(Child) = An_Expression) and then
                      ((Asis.Elements.Expression_Kind(Child) = An_Identifier) or
                       (Asis.Elements.Expression_Kind(Child) = A_Selected_Component))) then
                     if ((not Param_Tree_Empty) and then
                         (Get_Last_Param_Node.all in Nodes.Param_Alloc_Node_Object)) then
                        -- The current Node on the P-Stack is an Allocation Node!
                        Nodes.Add_Type_Item(Nodes.Param_Alloc_Node_Object(Get_Last_Param_Node.all),
                                            UStr_List.Get_Head(Trans_Data.Names));
                     elsif (Type_Ptr /= null) then
                        -- We have a Type Description Structure so save the Supertype
                        if (UStr_List.Empty(Type_Ptr.Super_Type)) then
                           Type_Ptr.Super_Type := Tokenize(U2s(UStr_List.Get_Head(Trans_Data.Names)), ".");
                        else
                           Put_Line(Error, "[Transformation] [ERR]" &
                                    " Transformation Logic Bug: Several Supertypes found!");
                        end if;
                     elsif (not Trans_Data.Ignore_Mode) then
                        Put_Line(Error, "[Transformation] [ERR]" &
                                 " No Type Structure available for Subtype Description!");
                     end if;
                  end if;

                  -- Attempt to save the Identifiers
                  Save_Identifiers (Trans_Data.Names);
                  UStr_List.Clear(Trans_Data.Names);

               when A_Component_Definition =>

                  -- Info: ** Component Definitions (HCHF)
                  --
                  --       A Component Definition means that we are dealing with the
                  --       Name of the Type of a an Item in an aggregate Structure;
                  --       This can be the Base Type for an Array, or it can be the Type
                  --       of an Entity inside a Record;
                  --       The only Child of this Node should be a Subtype Indication;
                  --
                  --       Since a Subtype Indication writes itself into the Supertype Field
                  --       before the Component Definition is processed, we will remove it from
                  --       there and write it into its rightful Place - the Component Type;
                  --
                  --       IMPORTANT NOTE: Please Note the following Inconsistency (?) in ASIS:
                  --
                  --        a) In a Variable Declaration (e.g. "I: Integer") the Type "Integer"
                  --           is saved as a direct Subtype Indication of the Variable, thus as
                  --           Supertype for the Variable itself;
                  --        b) In a Record Definition however, for a Component Declaration (e.g.
                  --           "record I: Integer; end record") the Type "Integer" is saved as
                  --           Component Definition -> Subtype Indication; This would mean we
                  --           are dealing with an aggregate "I" and is also the exact same
                  --           Structure as for an Array Declaration (e.g. "array () of Integer")
                  --           where "Integer" is also saved as CompDef/SubtypeInd;
                  --
                  --        This would mean we have a light Inconsistency in our CFG;
                  --         (+) An ordinary Variable has its Type saved in the Super Type Field
                  --         (+) An aggregate Array Variable has its Base Type saved in the
                  --             Component Type Field
                  --         (-) whereas a Component Class Variable has its Type saved in the
                  --             Component Type Field ALTHOUGH (!) it is NOT an aggregate Type;
                  --        We can be sure that the Component Class Variable is *NOT* an
                  --        aggregate Type, because in Records no anonymous Types are allowed!
                  --
                  --        Just to be sure, save any Identifiers found on the Way;
                  if ((Asis.Elements.Element_Kind(Child) = A_Definition) and then
                      (Asis.Elements.Definition_Kind(Child) = A_Subtype_Indication)) then
                     if (Type_Ptr /= null) then
                        if (not UStr_List.Empty(Type_Ptr.Super_Type)) then
                           -- Move the Data: Supertype -> Component Type
                           UStr_List.Duplicate(Type_Ptr.Super_Type, Type_Ptr.Component_Type);
                           -- Clear the Super Type Field
                           UStr_List.Clear(Type_Ptr.Super_Type);
                        else
                           Put_Line(Error, "[Transformation] [ERR]" &
                                 " Transformation Logic Bug: Supertype Structure empty!");
                        end if;
                     elsif (not Trans_Data.Ignore_Mode) then
                        Put_Line(Error, "[Transformation] [ERR]" &
                                 " No Type Structure available for Type Description!");
                     end if;
                  else
                     Put_Line(Warning, "[Transformation] [WARN]" &
                              " Child of Component Definition is not Subtype Indication!");
                  end if;

                  -- Attempt to save the Identifiers
                  Save_Identifiers (Trans_Data.Names);
                  UStr_List.Clear(Trans_Data.Names);

               when A_Record_Definition =>

                  -- Info: ** Record Definitions (HCHF)
                  --
                  --       The Children of the RecDefs, the Components, save themselves in the
                  --       PostOp Phase into the right Place, so we don't need to do anything
                  --       here;

                  null;

               when A_Task_Definition =>

                  -- Info: ** Task Definitions (POST)
                  --
                  --       No special Handling needed here;

                  null;

               when others =>
                  null;
            end case;

         when A_Declaration =>

            -- D E C L A R A T I O N S

            Declaration_Kind := Asis.Elements.Declaration_Kind (Parent);
            case Declaration_Kind is

               when An_Ordinary_Type_Declaration | A_Subtype_Declaration |
                 A_Private_Type_Declaration | A_Private_Extension_Declaration =>

                  -- Info: ** (Sub-)Type Declarations (HCHF)
                  --
                  --       The first Child (DefName/DefId) is the Name of the new Type;

                  if ((Asis.Elements.Element_Kind(Child) = A_Defining_Name) and then
                      (Asis.Elements.Defining_Name_Kind(Child) = A_Defining_Identifier)) then
                     -- This is the Name of the new Type - save it!
                     Trans_Data.Type_Decl.Type_Name := S2u(First_Line(Child));
                  end if;

               when A_Package_Declaration | A_Package_Body_Declaration |
                 A_Protected_Type_Declaration | A_Single_Protected_Declaration |
                 A_Protected_Body_Declaration |
                 A_Generic_Package_Declaration =>

                  -- Info: ** Flow Package Structure Declarations (HCHF)
                  --
                  --       These Structures have ONE Child of Type Defining Name / Defining
                  --       Identifier OR Defining Expanded Name, containing their Name;
                  --       As soon as we find such a Child we can create the associated Flow
                  --       Structure;
                  --
                  --       If the Structure is a Generic Package then we will have already
                  --       processed its Formal Declarations List - the Formals occur earlier
                  --       in the AST, so we only need to save the List into the newly created
                  --       Package;

                  -- Was this the Name of the Structure
                  if ((Asis.Elements.Element_Kind(Child) = A_Defining_Name) and then
                      ((Asis.Elements.Defining_Name_Kind(Child) = A_Defining_Identifier) or
                       (Asis.Elements.Defining_Name_Kind(Child) = A_Defining_Expanded_Name))) then
                     -- Yes -> Create a new Package
                     if (Declaration_Kind = A_Package_Declaration) then
                        New_Pkg_Ptr := new Pkgs.Simple_Spec_Object;
                     elsif (Declaration_Kind = A_Package_Body_Declaration) then
                        New_Pkg_Ptr := new Pkgs.Simple_Body_Object;
                     elsif (Declaration_Kind = A_Protected_Type_Declaration) then
                        New_Pkg_Ptr := new Pkgs.Prot_Type_Object;
                     elsif (Declaration_Kind = A_Single_Protected_Declaration) then
                        New_Pkg_Ptr := new Pkgs.Single_Prot_Object;
                     elsif (Declaration_Kind = A_Protected_Body_Declaration) then
                        New_Pkg_Ptr := new Pkgs.Prot_Object;
                     elsif (Declaration_Kind = A_Generic_Package_Declaration) then
                        New_Pkg_Ptr := new Pkgs.Gen_Object;
                     else
                        Put_Line(Error, "[Transformation] Unreachable assertion failed!");
                        New_Pkg_Ptr := null;
                     end if;
                     -- Check for the previous Structure
                     if Flow_Ptr /= null then
                        -- A previous Structure exists -> Set it as Father for the current one
                        Pkgs.Init(New_Pkg_Ptr.all, S2u(First_Line(Child)), Flow_Ptr, Parent);
                        Flow_Types.Add_Succ(Flow_Ptr.all, Flow_Types.Flow_Class_Ptr(New_Pkg_Ptr));
                     else
                        -- None - Must be the first Package -> Set a "null" Father and Clauses
                        Pkgs.Init(New_Pkg_Ptr.all, S2u(First_Line(Child)), null, Parent);
                        Flow_Types.Set_Refs(Flow_Types.Flow_Object(New_Pkg_Ptr.all), Trans_Data.Refs);
                        Flow_World.Add_Pkg(World, New_Pkg_Ptr);
                     end if;
                     -- Check if this is a Generic Package
                     if (Declaration_Kind = A_Generic_Package_Declaration) then
                        -- Yes -> Save the Formals into the new Structure
                        Flow_Types.Set_FDecls(Flow_Types.Flow_Object(New_Pkg_Ptr.all),
                                              Trans_Data.Form_Decls);
                        -- We no longer need these Formal Declarations
                        Flow_Types.Formal_Decl_Class_Ptr_List.Clear(Trans_Data.Form_Decls);
                     end if;
                     Put_Line (Debug, "[Transformation] New PKG: " & Pkgs.Get_Name(New_Pkg_Ptr.all));
                     -- Put the new Package onto the Flow Stack
                     Open_Flow(Flow_Types.Flow_Class_Ptr(New_Pkg_Ptr), Parent);
                     -- Update Variables
                     Get_Last_Flow_Info(Flow_Ptr, Pkg_Ptr, Cfg_Ptr);
                  end if;

               when A_Variable_Declaration | A_Component_Declaration |
                 A_Parameter_Specification | An_Entry_Index_Specification |
                 A_Discriminant_Specification | A_Loop_Parameter_Specification |
                 A_Constant_Declaration | A_Deferred_Constant_Declaration =>

                  -- Info: ** Variable-Class Declarations (HCHF)
                  --
                  --       These ASIS Types have a variable Number of Children, which are however
                  --       grouped in three Categories:
                  --         1. Names (Variables,Components,..) of ASIS Type Defining Name
                  --         2. Type of ASIS Type Definition OR Expression
                  --         3. Initialization of ASIS Type Expression
                  --
                  --       Since all these Types (Variable, Component, Parameter, Entry Index,..)
                  --       share a common Grammar, we are using a common Declaration Structure
                  --       to hold the Data until they are finished and we can insert them into
                  --       the final Flow Structure;
                  --
                  --       Note that we do not need to handle the Initialization here because we
                  --       have already created a Parameter Tree for this purpose in the PreOp
                  --       and saved it in the Declaration Structure; The Parameter Tree grows
                  --       automatically as more Function Calls and Parameters are being used;
                  --
                  --       Also note that we do not need to handle the Type of the Variable
                  --       either, beacause this happens automatically when processing the ASIS
                  --       Structures responsible for the Type Description; This however is only
                  --       Valid for Type-like ASIS Structures; Parameter Specifications use
                  --       Expression to define their Types - see below for more Details;
                  --
                  --       *Fifth Rider of the Apocalypse*: Each Declaration with a more complex
                  --       Type, packs it's Type Description in a Type Definition Subtree;
                  --       An Exception are, however, the Type Constraints; They are not
                  --       considered by ASIS to be a "more complex Type", and so, the Type
                  --       Specification consists only of a Subtype Indication + Constraints;
                  --       These Constraints are the Key to Pandora's Box, since they may contain
                  --       just about anything; For that, if we finish a Subtype Indication Child
                  --       for a Variable Class Declaration, we must move the entire Parameter
                  --       Tree created for the Initialization, in the first place, and convert
                  --       it to a Type Description P-Tree; After that we reinitialize the
                  --       Initialization P-Tree; This Action is legitimate since the Initialization
                  --       may only come AFTER the Type Description for the Variable;

                  if ((not Trans_Data.Ignore_Mode) and
                      (not ((Declaration_Kind = A_Parameter_Specification) and
                            (Trans_Data.Type_Decl /= null)))) then

                     Child_Element_Kind := Asis.Elements.Element_Kind (Child);
                     case Child_Element_Kind is

                        when A_Defining_Name =>
                           if (Asis.Elements.Defining_Name_Kind (Child) = A_Defining_Identifier) then
                              -- This is one of the new Entities being declared - add it to the List
                              UStr_List.Append(Trans_Data.Decl.Vars, S2u(First_Line(Child)));
                           end if;

                        when A_Definition =>
                           if ((Asis.Elements.Definition_Kind (Child) = A_Subtype_Indication)
                               and (Declaration_Kind /= A_Loop_Parameter_Specification)) then
                              -- This is the fifth Rider of the Apocalypse!
                              -- Move the Tree
                              Param_Node_Ptr := Nodes.Node_Object(Trans_Data.Decl.Init.all)'Access;
                              Nodes.Set_Info(Param_Node_Ptr.all, S2u("TYPE/DESCR"));
                              Trans_Data.Decl.VType.Descr_Tree :=
                                Flow_Types.Flow_Object(Param_Node_Ptr.all)'Access;
                              Close_Last_Param;
                              -- Create a new Tree
                              Param_Node_Ptr := new Nodes.Param_Node_Object;
                              Nodes.Init(Nodes.Param_Node_Object(Param_Node_Ptr.all), Flow_Ptr);
                              Nodes.Set_Info(Param_Node_Ptr.all, S2u("INIT/NODE"));
                              -- Save the Root of the Param Tree
                              Trans_Data.Decl.Init :=
                                Flow_Types.Flow_Object(Param_Node_Ptr.all)'Access;
                              Open_Param_Root(Param_Node_Ptr);
                           end if;

                        when An_Expression =>

                           -- Info: ** Expression in Variable & Co. Declarations
                           --
                           --       An Expression is usually the third Child of such a
                           --       Declaration and it holds the Initialization Code for the
                           --       Variable;
                           --
                           --       However: The Variable Type for a Parameter Specification is
                           --       NOT a Definition / Type (like in a Variable Declaration) but
                           --       also an Expression and thus we must treat it accordingly;
                           --
                           --       If this Child of a *_Specification is an Expression we must
                           --       check first to see if we already have a Type; Should
                           --       this not be the Case, we must save this Expression as the
                           --       Type of the Entity;
                           --
                           --       If this Expression is the Initialization then most of the
                           --       Processing will have already been done automatically;
                           --       However: if the Expression consists of only one (1) Variable,
                           --       this Variable will go by unnoticed;
                           --       Why: During Identifier Harvesting, the Names get added to the
                           --       last _Statement_ Node; A Declaration usually has no Statement
                           --       surrounding it, and an Attempt to add the Identifier to a
                           --       Statement will fail; We will explicitly try to add the
                           --       Variable to the current Parameter Node to prevent that;

                           if not UStr_List.Empty(Trans_Data.Names) then
                              if (((Declaration_Kind = A_Parameter_Specification) or
                                   (Declaration_Kind = A_Discriminant_Specification) or
                                   (Declaration_Kind = An_Entry_Index_Specification)) and
                                  (UStr_List.Empty(Trans_Data.Decl.VType.Super_Type))) then
                                 -- This Expression is the Type of a Parameter Specification
                                 Full_Name := UStr_List.Get_Head(Trans_Data.Names);
                                 Trans_Data.Decl.VType.Super_Type := Tokenize(U2s(Full_Name), ".");
                                 -- Make Room for new Names
                                 UStr_List.Clear(Trans_Data.Names);
                              elsif (Nodes.Get_Var_Name(Nodes.Param_Node_Object(Get_Last_Param_Node.all)) = "") then
                                 -- This Expression is the Initialization of the Variable & Co.
                                 Nodes.Add_Var_Item(Nodes.Param_Node_Object(Get_Last_Param_Node.all),
                                                    UStr_List.Get_Head(Trans_Data.Names));
                                 UStr_List.Clear(Trans_Data.Names);
                              end if;
                           end if;

                        when others =>
                           null;
                     end case;

                  end if;

               when A_Procedure_Declaration | A_Function_Declaration |
                 An_Entry_Declaration |
                 A_Procedure_Body_Stub | A_Function_Body_Stub | A_Task_Body_Stub =>

                  -- Info: ** CFG-Class Spec Declarations (HCHF)
                  --
                  --       We are intentionately ignoring these ASIS Types, since the bodies of
                  --       these Subprograms will have to already be visible by the Time they are
                  --       used, and since we will have seen the Bodies, we will already posess
                  --       all the Information we need about them;

                  null;

               when A_Procedure_Body_Declaration | A_Function_Body_Declaration |
                 An_Entry_Body_Declaration |
                 A_Single_Task_Declaration | A_Task_Type_Declaration | A_Task_Body_Declaration |
                 A_Generic_Procedure_Declaration | A_Generic_Function_Declaration =>

                  -- Info: ** CFG-Class Body and Generic Declarations (HCHF)
                  --
                  --       These Structures have ONE Child of Type Defining Name
                  --       containing their Name; As soon as we find such a Child we
                  --       can create the associated Flow Structures;
                  --
                  --       Although the Generic Types Specified here are NO Bodies (but Specs) we
                  --       still need to create Flow Structure for them, because we need
                  --       Containers for their Formal Types in the Flow World;
                  --       If the Structures are Generic, then we will have already processed
                  --       their Formal Declarations List - the Formals occur earlier in the AST,
                  --       so we only need to save the List into the newly created CFG;
                  --
                  --       Entry Body Declarations have a "Guardian", i.e. an Expression that
                  --       decides whether the CFG will run or not; Such Expressions outside
                  --       of Statements and Declarations are somewhat awkward to handle; We will
                  --       save the Root of this tree into the PTree Variable of the Flow Object;

                  -- Was this the Name of the CFG?
                  if (Asis.Elements.Element_Kind(Child) = A_Defining_Name) then
                     -- Yes -> Create a new CFG
                     -- Check if a Parent Structure exists, first
                     if Flow_Ptr = null then
                        -- No Parent, create a Default Package and append it to the World
                        New_Pkg_Ptr := new Pkgs.Def_Object;
                        Pkgs.Init(New_Pkg_Ptr.all, W2u("Default_PKG"), null, Parent);
                        Flow_Types.Set_Refs(Flow_Types.Flow_Object(New_Pkg_Ptr.all), Trans_Data.Refs);
                        Put_Line (Debug,"[Transformation] New PKG: " & Pkgs.Get_Name(New_Pkg_Ptr.all));
                        Flow_World.Add_Pkg(World, New_Pkg_Ptr);
                        -- Put the new Package onto the Flow Stack
                        Open_Flow(Flow_Types.Flow_Class_Ptr(New_Pkg_Ptr), Parent);
                        -- Update Variables
                        Get_Last_Flow_Info(Flow_Ptr, Pkg_Ptr, Cfg_Ptr);
                     end if;
                     -- Create the CFG
                     if (Declaration_Kind = A_Procedure_Body_Declaration) then
                        New_Cfg_Ptr := new CFGs.Proc_Object;
                     elsif (Declaration_Kind = A_Function_Body_Declaration) then
                        New_Cfg_Ptr := new CFGs.Func_Object;
                     elsif (Declaration_Kind = A_Task_Body_Declaration) then
                        New_Cfg_Ptr := new CFGs.Task_Object;
                     elsif (Declaration_Kind = An_Entry_Body_Declaration) then
                        New_Cfg_Ptr := new CFGs.Entry_Object;
                     elsif ((Declaration_Kind = A_Generic_Procedure_Declaration) or
                            (Declaration_Kind = A_Generic_Function_Declaration)) then
                        New_Cfg_Ptr := new CFGs.Generic_Object;
                     elsif (Declaration_Kind = A_Single_Task_Declaration) then
                        New_Cfg_Ptr := new CFGs.Single_Task_Object;
                     elsif (Declaration_Kind = A_Task_Type_Declaration) then
                        New_Cfg_Ptr := new CFGs.Task_Type_Object;
                     else
                        New_Cfg_Ptr := null;
                        Put_Line(Error, "[Transformation] Unreachable assertion failed!");
                     end if;
                     -- Initialize Structures
                     CFGs.Init(New_Cfg_Ptr.all, S2u(First_Line(Child)), Flow_Ptr, Parent);
                     Flow_Types.Add_Succ(Flow_Ptr.all, Flow_Types.Flow_Class_Ptr(New_Cfg_Ptr));
                     -- If this is the Beginning of a generic Structure we must save the
                     -- formal Declarations gathered so far
                     if ((Declaration_Kind = A_Generic_Procedure_Declaration) or
                         (Declaration_Kind = A_Generic_Function_Declaration)) then
                        Flow_Types.Set_FDecls(Flow_Types.Flow_Object(New_Cfg_Ptr.all), Trans_Data.Form_Decls);
                        -- We no longer need the Formal Declarations saved in the List
                        Flow_Types.Formal_Decl_Class_Ptr_List.Clear(Trans_Data.Form_Decls);
                     end if;
                     Put_Line (Debug, "[Transformation] New CFG: " & CFGs.Get_Name(New_Cfg_Ptr.all));
                     -- Put the new Package onto the Flow Stack
                     Open_Flow(Flow_Types.Flow_Class_Ptr(New_Cfg_Ptr), Parent);
                     -- Update Variables
                     Get_Last_Flow_Info(Flow_Ptr, Pkg_Ptr, Cfg_Ptr);
                  elsif ((Declaration_Kind = An_Entry_Body_Declaration) and
                         (Asis.Elements.Element_Kind(Child) = An_Expression)) then

                     -- This is the "Guardian" of an Entry Body; Check for a PTree
                     if (Get_Last_Param_Root /= null ) then
                        Flow_Types.Set_PTree(Flow_Ptr.all, Flow_Types.Flow_Object(Get_Last_Param_Root.all)'Access);
                     elsif not UStr_List.Empty(Trans_Data.Names) then
                        -- No PTree, but Names are available
                        -- Create a new Parameter Node and save the Identifiers in it
                        Node_Ptr := new Nodes.Param_Node_Object;
                        Nodes.Init(Nodes.Param_Node_Object(Node_Ptr.all), Flow_Ptr);
                        Nodes.Set_Info(Node_Ptr.all, S2u("Guardian"));
                        Nodes.Add_Var_Item(Nodes.Param_Node_Object(Node_Ptr.all),
                                           UStr_List.Get_Head(Trans_Data.Names));
                        Flow_Types.Set_PTree(Flow_Ptr.all, Flow_Types.Flow_Object(Node_Ptr.all)'Access);
                        UStr_List.Clear(Trans_Data.Names);
                     end if;

                  end if;

               when A_Formal_Object_Declaration | A_Formal_Type_Declaration |
                 A_Formal_Procedure_Declaration | A_Formal_Function_Declaration |
                 A_Formal_Package_Declaration | A_Formal_Package_Declaration_With_Box =>

                  -- Info: ** Formal Declarations (HCHF)
                  --
                  --       The Formal Declarations are very similar to the standard Variable &
                  --       & Co. Declarations; They have three Types of Children:
                  --         1. New Entity Names of ASIS Type Defining Name / Defining Identifier
                  --         2. The Old Enity Name / the Type of ASIS Type Expression
                  --         3. An Initialization Expression of ASIS Type Expression
                  --       Note that eventual Parameters are not Important here;
                  --
                  --       The Handling is straight-forward: if the Child that has just finished
                  --       is a Defining Identifier, add it to the Variables;
                  --
                  --       The Type must be saved by Hand, since the automatical Type Recognition
                  --       only keeps an Eye out for Type-Like ASIS Definitions, and the Type of
                  --       a Formal Declaration is saved as an Expression;
                  --
                  --       The Initializations do not need Handling here; It is done automatically;

                  -- Retreive the last Formal Declaration
                  Form_Decl := Flow_Types.Formal_Decl_Class_Ptr_List.Get_Tail(Trans_Data.Form_Decls);

                  Child_Element_Kind := Asis.Elements.Element_Kind (Child);
                  case Child_Element_Kind is

                     when A_Defining_Name =>
                        if (Asis.Elements.Defining_Name_Kind (Child) = A_Defining_Identifier) then
                           -- This is one of the new Names
                           UStr_List.Append(Form_Decl.Vars, S2u(First_Line(Child)));
                        end if;

                     when An_Expression =>
                        -- This is either the Type of the Declaration or the Initialization
                        -- Check if the Type has been set yet first!
                        if (not UStr_List.Empty(Trans_Data.Names)) then
                           if UStr_List.Count(Form_Decl.VType.Super_Type) = 0 then
                              -- This is indeed the Type
                              Full_Name := UStr_List.Get_Head(Trans_Data.Names);
                              Form_Decl.VType.Super_Type := Tokenize(U2s(Full_Name), ".");
                              -- Make Room for new Names
                              UStr_List.Clear(Trans_Data.Names);
                           else
                              -- The Type has already been set; Must be the Initialization
                              Nodes.Add_Var_Item(Nodes.Param_Node_Object(Get_Last_Param_Node.all),
                                                 UStr_List.Get_Head(Trans_Data.Names));
                              UStr_List.Clear(Trans_Data.Names);
                           end if;
                        end if;

                     when others =>
                        null;
                  end case;

               when A_Procedure_Renaming_Declaration | A_Function_Renaming_Declaration |
                 A_Package_Renaming_Declaration |
                 A_Generic_Procedure_Renaming_Declaration | A_Generic_Function_Renaming_Declaration |
                 A_Generic_Package_Renaming_Declaration |
                 An_Object_Renaming_Declaration | An_Exception_Renaming_Declaration  =>

                  -- Info: ** Entity Renaming Declarations (HCHF)
                  --
                  --       The first Child of a Renaming Declaration is the New Name and the last
                  --       one is the Old Name; Since we keep a List of Names at all Times, we
                  --       could save both these Names into the Renaming Structure when
                  --       PostOp-ing the Declaration, however it is more safe (we can be sure
                  --       no Names will get lost) if we save the New Name after the first Child
                  --       and the old one after the last (in PostOp);
                  --       The first Child also has to be a Defining Identifier;

                  if (Count = 1) then
                     Trans_Data.Rename.New_Name := S2u(First_Line(Child));
                  end if;

               when A_Package_Instantiation |
                 A_Procedure_Instantiation | A_Function_Instantiation =>

                  -- Info: ** Generic Instantiations (HCHF)
                  --
                  --       In a Generic Instantiation, the first Child is the Name of the new
                  --       Entity and the second Child is the Name of the old one;

                  if (Count = 1) then
                     -- Save the new Name
                     Trans_Data.Rename.New_Name := S2u(First_Line(Child));
                     -- Make Room for other Names
                     UStr_List.Clear(Trans_Data.Names);
                  elsif (Count = 2) then
                     if (UStr_List.Count(Trans_Data.Names) /= 1) then
                        Put_Line(Warning,"[Transformation] [WARN] " &
                                 "Too many Names for generic Entity Instantiation!");
                     end if;
                     -- Save the Name of the old (generic) Entity
                     Full_Name := UStr_List.Get_Head(Trans_Data.Names);
                     Trans_Data.Rename.Old_Name := Tokenize(U2s(Full_Name), ".");
                     -- Make Room for other Names
                     UStr_List.Clear(Trans_Data.Names);
                  end if;

               when others =>
                  null;
            end case;

         when An_Expression =>

            -- E X P R E S S I O N S

            Expression_Kind := Asis.Elements.Expression_Kind (Parent);
            case Expression_Kind is

                when A_Positional_Array_Aggregate | A_Record_Aggregate =>

                   -- Info: ** Aggregates (HCHF)
                   --
                   --       Nothing to do here;

                   null;

               when An_Attribute_Reference =>

                  -- Info: ** Attribute References (HCHF)
                  --
                  --       The ARs are being processed in the PostOp-Method, and there is nothing
                  --       to be done here;

                  null;

               when An_Allocation_From_Subtype =>

                  -- Info: ** Allocation From Subtype (HCHF)
                  --
                  --       Since the Subtree with the AFS as Root has the same Structure as a
                  --       Type Definition, it will be processed elsewhere and there is nothing
                  --       we ca do here;

                  null;

               when A_Type_Conversion =>

                  -- Info: ** Type Conversions (HCHF)
                  --
                  --       The Type Conversion is the typical Type Casting; It has two Children
                  --       of Type Expression/*; The first one is the Type we are convertion to,
                  --       and the second one is the Name of the Structure being converted;
                  --       We do not need the Target Type, but since it is an Expression/* is
                  --       will be saved into the Identifier List; We thusly need to flush this
                  --       List (without saving) before proceeding with the Name of the Variable;

                  if Count = 1 then
                     UStr_List.Clear(Trans_Data.Names);
                  elsif Count > 2 then
                     Put_Line(Warning, "[Transformation] [WARN] Transformation Logic Bug:" &
                              " Type Conversion has more than two Children!");
                  end if;

               when An_Identifier =>

                  -- Info: ** Identifiers (HCHF)
                  --
                  --       We save the Identifier in the PostOp and there is nothing to do here;

                  null;

               when An_Operator_Symbol =>

                  -- Info: ** Operator Symbols (HCHF)
                  --
                  --       We save the Operator in the PostOp and there is nothing to do here;

                  null;

               when A_Function_Call =>

                  -- Info: ** Function Calls (HCHF)
                  --
                  --       The first Child of a Function Call is the Name of the called Function;
                  --       We will also add all the found Identifiers to the active Node since
                  --       complex Structures (like Indexed Components, ..) may have been used;

                  if (Count=1) then
                     -- Add the Name of the called Function to the Call Node
                     Nodes.Add_Call_Item(Nodes.Call_Node_Object(Get_Last_Param_Node.all),
                                         S2u(First_Line(Child)));
                     -- Write all found Identifiers to the last Node and clear the List
                     Save_Identifiers(Trans_Data.Names);
                     UStr_List.Clear(Trans_Data.Names);
                  end if;

               when A_Selected_Component =>

                  -- Info: ** Selected Components (HCHF)
                  --
                  --       For easier Handling of the entire Name we will append all found
                  --       Children of the Selected Component back together and thusly
                  --       reconstruct the original Name; We can safely retokenize this Name
                  --       at a later Point in Time;
                  --
                  --       A Selected Component has two Children; The first one is an Expression
                  --       / Identifier, and the second one is the Root to a recursive Subtree
                  --       containing the rest of the Name in a similar Hierarchy;
                  --
                  --       We are appending the Names back together when going back upwards
                  --       (bottom up); This way, the right Subtree is already compressed in a
                  --       "Name{i+1}.Name{i+2}..Name{n}"-Form; We now append the compressed #
                  --       Names from this Subtree to the Name of the left Child "Name{i}";
                  --       When the Transformation has finished, the compressed Name will lie
                  --       on the first Position in the Names List;
                  --
                  --       We use the Name List as a Stack here, because it is easier to
                  --       concatenate the last two Strings this way;

                  if (Count = 2) and (UStr_List.Count(Trans_Data.Names) > 1) then
                     UStr_List.Pop(Trans_Data.Names, Full_Name);
                     UStr_List.Pop(Trans_Data.Names, Component_Name);
                     UStr_List.Push(Trans_Data.Names, Component_Name & S2u(".") & Full_Name);
                  end if;

               when An_Indexed_Component | A_Slice =>

                  -- Info: ** Indexed (-Class) Components (HCHF)
                  --
                  --       An Indexed Component can have a complex Param Tree and its Name may
                  --       become lost on the way, especially if Function Calls are involved, so
                  --       we force saving the Name of said Variable in the last Param Node
                  --       (if any available);

                  if ((Count = 1) and (not UStr_List.Empty(Trans_Data.Names)) and
                      ((Get_Last_Param_Node /= null) and then
                       (Get_Last_Param_Node.all in Nodes.Param_Node_Object))) then
                     -- Save the Name of the Variable
                     if (Nodes.Get_Var_Name(Nodes.Param_Node_Object(Get_Last_Param_Node.all)) = "") then
                        Nodes.Add_Var_Item(Nodes.Param_Node_Object(Get_Last_Param_Node.all),
                                           UStr_List.Get_Head(Trans_Data.Names));
                        -- Check if this is a Renaming
                        if ((Trans_Data.Rename /= null) and then
                            (Trans_Data.Rename.all in Flow_Types.Obj_Alias)) then
                           Trans_Data.Rename.Old_Name := Tokenize(U2s(UStr_List.Get_Head(Trans_Data.Names)),".");
                        end if;
                     else
                        Put_Line(Warning, "[Transformation] [WARN]" &
                                 " Found Complex Structure Name! Dropping Token <" &
                                 U2s(UStr_List.Get_Head(Trans_Data.Names)) & "> for Name" &
                                 " <" & Nodes.Get_Var_Name(Nodes.Param_Node_Object(Get_Last_Param_Node.all)) &
                                 ">!");
                     end if;
                     -- Add the found Identifiers to the last Node and clear the List
                     Save_Identifiers(Trans_Data.Names);
                     UStr_List.Clear(Trans_Data.Names);
                  end if;

               when others =>
                  null;
            end case;

         when A_Path =>

            -- P A T H S

            -- Info: ** Paths (HCHF)
            --
            --       This Trigger does not bear much Information; We only know that a Child of
            --       a Path (some Statement) has finished; There is noting we can do here;

            -- Save Identifiers
            Save_Identifiers(Trans_Data.Names);

         when A_Statement =>

            -- S T A T E M E N T S

            -- ** GENERAL STATEMENT (HCHF) HANDLING

            -- Info: ** Statements (HCHF)
            --
            --       Statements have a rather variable Tree Syntax which makes them relatively
            --       complicated to analyse; A Statement can have several Labels of ASIS Type
            --       Defining Name / Defining Identifier;
            --
            --       Our first Action is to save the found Identifiers; This may create false
            --       Warnings especially when a new CFG has just been placed on the Stack and
            --       there is no current Node available.
            --
            --       Our second Action is to isolate the Names of the Statement;

            -- Save the Identifiers found so far into the Statement
            Save_Identifiers(Trans_Data.Names);

            Child_Element_Kind := Asis.Elements.Element_Kind (Child);
            if Child_Element_Kind = A_Defining_Name then
               Node_Ptr := Get_Last_Node;
               if (Node_Ptr /= null) then
                  -- The Node is NULL if a CFG has just been put on the Stack; this happens
                  -- e.g. with named Blocks; Add the new Name to the List of existing Names;
                  if (String'(Nodes.Get_Name(Node_Ptr.all)) = "") then
                     Nodes.Set_Name(Node_Ptr.all, S2u(First_Line(Child)));
                  else
                     Nodes.Set_Name(Node_Ptr.all, S2u(Nodes.Get_Name(Node_Ptr.all)
                                                      & "," & First_Line(Child)));
                  end if;
               end if;
            end if;

            -- ** SPECIAL STATEMENT (HCHF) HANDLING

            -- Info: ** Statements (HCHF)

            Statement_Kind := Asis.Elements.Statement_Kind (Parent);
            case Statement_Kind is

               when An_Assignment_Statement =>

                  -- Info: ** Assignment Statements (HCHF)
                  --
                  --       The Children-Structure of an Assignment Statement is
                  --       <Def/Name>* <Expr/?> <Expr/?>; When terminating an Expression Child,
                  --       it is difficult to say which is which especially if DefNames are
                  --       present; However, DefNames do not occur in the Transformation Data
                  --       Container's Name List and since the Name List is cleared when
                  --       beginning with a Statement, we can use per Default the first found
                  --       Identifiers as Variable Names;
                  --
                  --       We usually don't care much about where a Variable Name occurrs in the
                  --       final CFG, as long as it occurrs at all! In this case however we must
                  --       make sure we save the leading Variable of the Left Side into the Root
                  --       Node of the LHS Tree, because THIS Variable is the one that will be
                  --       written;
                  --
                  --       So. The Default Action is to write the first Identifier of the Name
                  --       List into the current Node of the Parameter Tree;
                  --
                  --       LHS: The LHS can only contain a Variable (the one to be written);
                  --       This Variable will be copied into the Root Node of the LHS Tree;
                  --       After this, the Trees will be switched - the RHS Tree becomes active;
                  --       In Case of Indexed Components, the Name of the Indexed Component has
                  --       already been copied into the Param Node (during the Handling of the
                  --       IC itself);
                  --
                  --       RHS: If the RHS also consists of only one Variable, this Variable will
                  --       still be in the List and we can copy it into said Tree; If it was a
                  --       Function, the List will be empty; In any Case, the Default Action is
                  --       correct;
                  --
                  --       The second Action for the ASs is to set the LHS/RHS Trees correctly;
                  --       As soon as the LHS of the AS has finished, we must initialize the
                  --       RHS Tree;

                  Child_Element_Kind := Asis.Elements.Element_Kind (Child);
                  if Child_Element_Kind = An_Expression then

                     -- This is an Expression - i.e. the LHS or the RHS
                     -- Default Action: Try to save a Variable into the current Parameter Node
                     if ((not UStr_List.Empty(Trans_Data.Names)) and
                         (Nodes.Get_Var_Name(Nodes.Param_Node_Object(Get_Last_Param_Node.all)) = "")) then
                        -- We have Names and this Node doesn't contain any Variables yet!
                        -- Write the Variable into the current Param Tree Node
                        Nodes.Add_Var_Item(Nodes.Param_Node_Object(Get_Last_Param_Node.all),
                                           UStr_List.Get_Head(Trans_Data.Names));
                     end if;

                     -- Secondary Action: Switch the Parameter Trees
                     if (Nodes.Get_LHS_Tree(Nodes.Assign_Node_Object(Get_Last_Node.all)) =
                         Nodes.Get_RHS_Tree(Get_Last_Node.all)) then
                        -- We don't have a RHS Tree yet - switch Trees
                        -- Close the LHS Tree
                        Close_All_Params;
                        -- Initialize new RHS Tree
                        Param_Node_Ptr := new Nodes.Param_Node_Object;
                        Nodes.Init(Nodes.Param_Node_Object(Param_Node_Ptr.all), Flow_Ptr);
                        Nodes.Set_RHS_Tree(Get_Last_Node.all, Param_Node_Ptr);
                        Nodes.Set_Info(Param_Node_Ptr.all, S2u("ASSIGN/RHS"));
                        -- Open the Tree
                        Open_Param_Root(Param_Node_Ptr);
                     end if;

                  end if;

               when A_Goto_Statement =>

                  -- Info: ** Jump-Class Statement / GOTO (HCHF)
                  --
                  --       Goto Statements have an Expression/Identifier as Target;
                  --       We will save this Name;

                  Jump_Node_Ptr := Get_Last_Node;
                  Nodes.Set_Target(Nodes.Jump_Node_Object(Jump_Node_Ptr.all),
                                   S2u(First_Line(Child)));
                  UStr_List.Clear(Trans_Data.Names);

               when An_Exit_Statement =>

                  -- Info: ** Jump-Class Statement / EXIT (HCHF)
                  --
                  --       Exit Class Statements have an Expression/Identifier as Target;
                  --       We will save this Name;
                  --       Possible scenarios:
                  --        * exit;
                  --        * exit Target; (!)
                  --        * exit when Variable; (!)
                  --        * exit Target wen Variable; (!)
                  --        + the last two with complex Expressions
                  --        + all the above with Statement Labels
                  --       We only need to identify the Target here (Expr/Id); The (!)-marked
                  --       Cases have such Children;

                  Jump_Node_Ptr := Get_Last_Node;
                  Child_Element_Kind := Asis.Elements.Element_Kind (Child);
                  if ((Child_Element_Kind = An_Expression) and then
                        (Asis.Elements.Expression_Kind (Child) = An_Identifier)) then
                     -- Okay, we found an Expr/Id child;
                     -- Does this Exit node have a target at all?
                     if (Asis.Elements.Expression_Kind (Asis.Statements.Exit_Loop_Name(Parent)) /= Not_An_Expression) then
                        -- Yes, this node does have a loop name as exit target
                        -- Did we save it yet?
                        if (Nodes.Get_Target(Nodes.Jump_Node_Object(Jump_Node_Ptr.all)) = S2u("")) then
                           -- Nope, we didn't save it yet - save now!
                           Nodes.Set_Target(Nodes.Jump_Node_Object(Jump_Node_Ptr.all),
                                            S2u(First_Line(Child)));
                           UStr_List.Clear(Trans_Data.Names);
                        end if;
                     end if;
                  end if;

               when A_Block_Statement =>

                  -- Info: ** Block Statements (HCHF)
                  --
                  --       This Trigger only tells us that a Statement has finished; There is
                  --       nothing we can do about it;

                  null;

               when An_Accept_Statement =>

                  -- Info: ** Accept Statements (HCHF)
                  --
                  --       An Accept can consist of an atomic Action, OR it can build up an
                  --       entire CFG with Parameters, Declarations, Statements, a.s.o.
                  --
                  --       To be fit for the latter, we created in the PreOp (just like for the
                  --       Block Statements) a new Call-Class Node and a new CFG to hold all the
                  --       Accept-internal Information (if any); The Call Node and the new CFG
                  --       have been linked only through the Node's Id: the CFG has it as Name;
                  --
                  --       Since the really important Thing, however, is the Name of the Entry,
                  --       we will store this Name into the CFG's Info Variable, as soon as we
                  --       find it; The Name is an Expression/Identifier and should be easy
                  --       enough to find;

                  if ((Asis.Elements.Element_Kind(Child) = An_Expression) and then
                      (Asis.Elements.Expression_Kind(Child) = An_Identifier)) then
                     CFGs.Set_Name(Cfg_Ptr.all, S2u(First_Line(Child)));
                  end if;

               when A_Procedure_Call_Statement | A_Raise_Statement | An_Entry_Call_Statement |
                 A_Requeue_Statement | A_Requeue_Statement_With_Abort | An_Abort_Statement =>

                  -- Info: ** Call-Class Statements Statements (HCHF)
                  --
                  --       Call-Class Statements have Expression/Identifier Children with the
                  --       Names of the called Methods, raised Exceptions, ..;
                  --       We will save this Name as the Call's Destination;

                  if (Asis.Elements.Element_Kind(Child) = An_Expression) then
                     if ((Asis.Elements.Expression_Kind(Child) = An_Identifier) or
                         (Asis.Elements.Expression_Kind(Child) = A_Selected_Component) or
                         (Asis.Elements.Expression_Kind(Child) = An_Indexed_Component)) then
                        -- This is the Target -> Save it into the Node (Origianl AND Parameter)
                        Call_Node_Ptr := Get_Last_Node;
                        Param_Node_Ptr := Get_Last_Param_Node;
                        if (not UStr_List.Empty(Trans_Data.Names)) then
                           Nodes.Add_Call_Item(Nodes.Call_Node_Object(Call_Node_Ptr.all),
                                               UStr_List.Get_Head(Trans_Data.Names));
                           Nodes.Add_Call_Item(Nodes.Call_Node_Object(Param_Node_Ptr.all),
                                               UStr_List.Get_Head(Trans_Data.Names));
                        end if;
                     end if;
                  end if;

               when An_If_Statement | A_Case_Statement | A_Selective_Accept_Statement |
                 A_Conditional_Entry_Call_Statement | A_Timed_Entry_Call_Statement |
                 An_Asynchronous_Select_Statement =>

                  -- Info: ** Branch-Class Nodes (HCHF)
                  --
                  --       We have reached the End of a Branch and the Top of the Node Stack has
                  --       the following Structure "If-EndIf-Node"; Before the Node of the next
                  --       Branch can be processed we must restore the "If-EndIf-If" State;
                  --
                  --       During the Process we also have to link the Node we are removing to
                  --       to "EndIf"; We need however to check if the Node we have just
                  --       removed isn't the "If" Node itself; This can happen with Case
                  --       Statements, where the first Child is the Test Variable and thusly no
                  --       actual CFG Node (the Stack State didn't change);
                  --
                  --       Another Check is for the Else Path; At the End of each Branch we must
                  --       check if this was an Else Path; Should we find no Else Alternative,
                  --       we will add a Default Else Path (jumping over the "If" Body) at the
                  --       End of the "If" Statement;

                  -- Remove the last Node of the Branch
                  Node_Ptr := Remove_Last_Node;
                  -- Find the "If" Node
                  Ei_Node_Ptr := Remove_Last_Node;
                  If_Node_Ptr := Get_Last_Node;
                  -- Restore the "If"-"EndIf"-"If" Order on the Stack
                  Add_Node(Ei_Node_Ptr);
                  Add_Node(If_Node_Ptr);

                  if (Node_Ptr /= If_Node_Ptr) then
                     -- Link the last Node of the Branch to the "EndIf"
                     Nodes.Add_Succ(Node_Ptr.all, Flow_Types.Flow_Class_Ptr(Ei_Node_Ptr));
                     -- Link the "EndIf" backwards to the last Node of the Branch
                     Nodes.Add_Pred(Ei_Node_Ptr.all, Flow_Types.Flow_Class_Ptr(Node_Ptr));
                     -- Check for the Else Path
                     if (Asis.Elements.Path_Kind (Child) = An_Else_Path) then
                        -- The Mark meaning "If has an Else Path" is "2"
                        Nodes.Set_Mark(If_Node_Ptr.all, 2);
                     end if;
                  end if;

               when A_Loop_Statement | A_While_Loop_Statement | A_For_Loop_Statement =>

                  -- Info: ** Loop-Class Nodes (HCHF)
                  --
                  --       A Child of the Loop Statement has finished; This Child can either
                  --       hold Information about the Loop Control or be some Statement in the
                  --       Loop itself; The Loop Control Information will be saved part in the
                  --       Loop Node part in the Parameter Tree and can be ignored here; The
                  --       Statements in the Loop also bear no Interes for us here;

                  null;

               when A_Return_Statement | A_Terminate_Alternative_Statement =>

                  -- Info: ** Termination-Class Nodes (HCHF)
                  --
                  --       Information about the returned Value will be held part in the CFG Node
                  --       part in the Parameter Tree, so there is nothing we can do here;

                  null;

               when others =>
                  null;
            end case;

            -- Make Room for new Names
            UStr_List.Clear(Trans_Data.Names);

         when An_Exception_Handler =>

            -- E X C E P T I O N S

            -- Info: ** Exception Handlers (HCHF)
            --
            --       The Name of the Exception being caught is of ASIS Type Expression /
            --       Identifier; As soon as we have it, we can create a CFG Structure for the
            --       Statements executed by the Handler; We are using the Child Number here
            --       because the "Others Choice" is not an Expression but a Definition;
            --

            -- Was this the Name of the CFG?
            if (Count = 1) then
               -- Yes -> Create a new CFG
               New_Cfg_Ptr := new CFGs.Except_Object;
               CFGs.Init(New_Cfg_Ptr.all, S2u(First_Line(Child)), Flow_Ptr, Parent);
               Put_Line (Debug, "[Transformation] New CFG: " & CFGs.Get_Name(New_Cfg_Ptr.all));
               CFGs.Add_Succ(Cfg_Ptr.all, Flow_Types.Flow_Class_Ptr(New_Cfg_Ptr));
               -- Put the new Package onto the Flow Stack
               Open_Flow(Flow_Types.Flow_Class_Ptr(New_Cfg_Ptr), Parent);
               -- Update the Variables
               Get_Last_Flow_Info(Flow_Ptr, Pkg_Ptr, Cfg_Ptr);
            end if;

         when others =>
            null;
      end case;

   end Handle_Child_Finish;

   -- ** U N I T   T R A N S F O R M A T I O N   H A N D L E R S

   procedure Handle_Unit_Started(Name: String; Class: Asis.Unit_Classes) is
   begin
      Put_Line (Debug, "[Transformation] Started transformation of " & Name);
   end Handle_Unit_Started;

   procedure Handle_Unit_Finished(Name: String; Class: Asis.Unit_Classes) is
      Flow_Ptr: Flow_Types.Flow_Class_Ptr := null;
      Pkg_Ptr:  Pkgs.Pkg_Class_Ptr        := null;
      Cfg_Ptr:  CFGs.CFG_Class_Ptr        := null;
   begin

      -- Info: ** Unit-Has-Finished Handler (PostOp)
      --
      --       Each Flow Structure (Pkg/CFG) is being closed together with its generator Element;
      --       In other words, if the Program finishes correctly no Flow Structures or Elements
      --       are allowed to remain open once the Unit finishes;
      --
      --       The only Exception from this Rule is the Default Package, that has NO Generator
      --       Element; We have to perform an extra Check for this Package;

      -- Identify the current Structures
      Get_Last_Flow_Info(Flow_Ptr, Pkg_Ptr, Cfg_Ptr);

      Put_Line(Debug, "[Transformation] Completed transformation of " & Name);

      -- Check the Element Stack
      if (not Asis.Elements.Is_Equal(Get_Last_Element, Nil_Element)) then
         Put_Line (Debug, "[Transformation] Element stack NOT empty when Unit finished!");
         -- Purge the Stack
         Close_All_Elements;
      end if;

      -- Check the Param Stack
      if (not Param_Tree_Empty) then
         Put_Line (Warning, "[Transformation] [WARN] Param Stack NOT empty when Unit finished!");
         -- Purge the Stack
         Close_All_Params;
      end if;

      -- Check the Flow Stack
      if (Flow_Ptr /= null) then
         if (Pkg_Ptr /= null) and then (Pkg_Ptr.all in Pkgs.Def_Object) then
            -- This is the Default Package -> close it!
            Put_Line(Debug, "[Transformation] Closing default package.");
            Close_Last_Flow;
            Get_Last_Flow_Info(Flow_Ptr, Pkg_Ptr, Cfg_Ptr);
         end if;
         -- Try again: check the Flow Stack
         if (Flow_Ptr /= null) then
            -- Something's very wrong: Flow Structures are still Open!
            Put_Line (Error, "[Transformation] [ERR] Flow stack NOT empty when Unit finished!");
            -- Purge the Stack
            Close_All_Flows;
         end if;
      end if;

      -- Clear the Transformation Data Container

      Trans_Data.Decl := null;
      Flow_Types.Formal_Decl_Class_Ptr_List.Clear(Trans_Data.Form_Decls);
      Trans_Data.Ignore_Mode := False;
      UStr_List.Clear(Trans_Data.Names);
      Trans_Data.Rename := null;
      Flow_Types.Ref_Class_Ptr_List.Clear(Trans_Data.Refs);
      Trans_Data.Type_Decl := null;
      Trans_Data.Last_PTree := null;

   end Handle_Unit_Finished;

   -- ** T R A N S F O R M A T I O N   S T R U C T U R E   M E T H O D S

   -- * Element Stack related Methods

   procedure Open_Element (Element: Asis.Element) is
      Elem_Ptr: Stack_Element_Ptr := null;
   begin
      Elem_Ptr := new Stack_Element;
      Elem_Ptr.Elem := Element;
      Elem_Ptr.Child := 0;
      Element_Stack.Push(EStack, Elem_Ptr);
   end Open_Element;

   procedure Close_Last_Element is
      Elem_Ptr: Stack_Element_Ptr := null;
   begin
      if not Element_Stack.Empty(EStack) then
         Element_Stack.Pop(EStack, Elem_Ptr);
         if (Elem_Ptr /= null) then
            Free_Stack_Elem(Elem_Ptr);
         end if;
      end if;
   end Close_Last_Element;

   procedure Close_All_Elements is
   begin
      Element_Stack.Clear(EStack);
   end Close_All_Elements;

   function Get_Last_Element_Container return Stack_Element_Ptr is
      Elem_Ptr: Stack_Element_Ptr := null;
   begin
      if (not Element_Stack.Empty(EStack)) then
         Element_Stack.Peek(EStack, Elem_Ptr);
      end if;
      return Elem_Ptr;
   end Get_Last_Element_Container;

   function Get_Last_Element return Asis.Element is
      Elem_Ptr: Stack_Element_Ptr := Get_Last_Element_Container;
   begin
      if (Elem_Ptr /= null) then
         return Elem_Ptr.Elem;
      else
         return Nil_Element;
      end if;
   end Get_Last_Element;

   function Get_Last_Child_Nr return Integer is
      Elem_Ptr: Stack_Element_Ptr := Get_Last_Element_Container;
   begin
      if (Elem_Ptr /= null) then
         return Elem_Ptr.Child;
      else
         return -1;
      end if;
   end Get_Last_Child_Nr;

   procedure Set_Child_Finished is
      Elem_Ptr: Stack_Element_Ptr := Get_Last_Element_Container;
   begin
      if (Elem_Ptr /= null) then
         Elem_Ptr.Child := Elem_Ptr.Child + 1;
      end if;
   end Set_Child_Finished;

   -- * Flow Stack related Methods

   procedure Open_Flow (Flow_Ptr: in Flow_Types.Flow_Class_Ptr;
                        Elem:     in Asis.Element) is
      Flow_Stack_Elem_Ptr: Flow_Stack_Element_Ptr := null;
   begin
      Flow_Stack_Elem_Ptr := new Flow_Stack_Element;
      Flow_Stack_Elem_Ptr.Flow := Flow_Ptr;
      Flow_Stack_Elem_Ptr.Elem := Elem;
      Node_Stack.Init(Flow_Stack_Elem_Ptr.NStack);
      Flow_Stack.Push(FStack, Flow_Stack_Elem_Ptr);
   end Open_Flow;

   procedure Close_Last_Flow is
      Flow_Stack_Elem_Ptr: Flow_Stack_Element_Ptr := null;
      Node_Ptr, New_Node_Ptr: Nodes.Node_Class_Ptr := null;
   begin
      if not Flow_Stack.Empty(FStack) then
         Flow_Stack.Pop(FStack, Flow_Stack_Elem_Ptr);
         -- We have just removed an element from the flow stack; This element
         -- could be a CFG or a package; Should the former be the case, we must
         -- also add a tail-node to it;
         if not Node_Stack.Empty(Flow_Stack_Elem_Ptr.NStack) then
            Node_Stack.Pop(Flow_Stack_Elem_Ptr.NStack, Node_Ptr);
            New_Node_Ptr := new Nodes.Node_Object;
            Nodes.Init(New_Node_Ptr.all, Flow_Stack_Elem_Ptr.Flow, Nil_Element);
            Nodes.Set_Info(New_Node_Ptr.all, S2u("CFG END"));
            -- Link the node
            Nodes.Add_Succ(Node_Ptr.all, Flow_Types.Flow_Class_Ptr(New_Node_Ptr));
            Nodes.Add_Pred(New_Node_Ptr.all, Flow_Types.Flow_Class_Ptr(Node_Ptr));
            -- Save it as tail for the CFG
            CFGs.Set_Tail(CFGs.CFG_Object(Flow_Stack_Elem_Ptr.Flow.all),New_Node_Ptr);
         end if;
         Free_Flow_Stack_Elem(Flow_Stack_Elem_Ptr);
      end if;
   end Close_Last_Flow;

   procedure Close_All_Flows is
   begin
      Flow_Stack.Clear(FStack);
   end Close_All_Flows;

   function Get_Last_Flow_Container return Flow_Stack_Element_Ptr is
      Flow_Stack_Elem_Ptr: Flow_Stack_Element_Ptr := null;
   begin
      if not Flow_Stack.Empty(FStack) then
         Flow_Stack.Peek(FStack, Flow_Stack_Elem_Ptr);
         return Flow_Stack_Elem_Ptr;
      end if;
      return null;
   end Get_Last_Flow_Container;

   -- Get Information about the last Flow
   procedure Get_Last_Flow_Info (Flow_Ptr: out Flow_Types.Flow_Class_Ptr;
                                 Pkg_Ptr:  out Pkgs.Pkg_Class_Ptr;
                                 Cfg_Ptr:  out CFGs.CFG_Class_Ptr) is
      Flow_Stack_Elem_Ptr: Flow_Stack_Element_Ptr := Get_Last_Flow_Container;
      Top_Flow_Ptr: Flow_Types.Flow_Class_Ptr     := null;
      Top_Pkg_Ptr:  Pkgs.Pkg_Class_Ptr            := null;
      Top_Cfg_Ptr:  CFGs.CFG_Class_Ptr            := null;
   begin
      -- Attempt find the last Flow and convert it to either CFG or Pkg
      if Flow_Stack_Elem_Ptr /= null then
         Top_Flow_Ptr := Flow_Stack_Elem_Ptr.Flow;
         if (Top_Flow_Ptr.all in CFGs.CFG_Object'Class) then
            Top_Cfg_Ptr := CFGs.CFG_Class_Ptr(Top_Flow_Ptr);
         elsif (Top_Flow_Ptr.all in Pkgs.Pkg_Object'Class) then
            Top_Pkg_Ptr := Pkgs.Pkg_Class_Ptr(Top_Flow_Ptr);
         else
            Put_Line(Error, "[Transformation] [ERR] Incorrect Flow Type on Flow Stack!");
            Top_Flow_Ptr := null;
         end if;
      end if;
      Flow_Ptr := Top_Flow_Ptr;
      Pkg_Ptr  := Top_Pkg_Ptr;
      Cfg_Ptr  := Top_Cfg_Ptr;
   end Get_Last_Flow_Info;

   function Get_Last_Flow_Elem return Asis.Element is
      Flow_Stack_Elem_Ptr: Flow_Stack_Element_Ptr := Get_Last_Flow_Container;
   begin
      if Flow_Stack_Elem_Ptr /= null then
         return Flow_Stack_Elem_Ptr.Elem;
      else
         return Nil_Element;
      end if;
   end Get_Last_Flow_Elem;

   -- * Node Stack related Methods

   procedure Add_Node (Node: in Nodes.Node_Class_Ptr) is
      Flow_Stack_Elem_Ptr: Flow_Stack_Element_Ptr := Get_Last_Flow_Container;
   begin
      if (Flow_Stack_Elem_Ptr /= null) then
         if (Flow_Stack_Elem_Ptr.Flow.all in CFGs.CFG_Object'Class) then
            Node_Stack.Push(Flow_Stack_Elem_Ptr.NStack, Node);
         end if;
      end if;
   end Add_Node;

   function Remove_Last_Node return Nodes.Node_Class_Ptr is
      Flow_Stack_Elem_Ptr: Flow_Stack_Element_Ptr := Get_Last_Flow_Container;
      Last_Node_Ptr:       Nodes.Node_Class_Ptr   := null;
   begin
      if (Flow_Stack_Elem_Ptr /= null) then
         if (Flow_Stack_Elem_Ptr.Flow.all in CFGs.CFG_Object'Class) then
            if not Node_Stack.Empty(Flow_Stack_Elem_Ptr.NStack) then
               Node_Stack.Pop(Flow_Stack_Elem_Ptr.NStack, Last_Node_Ptr);
            end if;
         end if;
      end if;
      return Last_Node_Ptr;
   end Remove_Last_Node;

   function Get_Last_Node return Nodes.Node_Class_Ptr is
      Flow_Stack_Elem_Ptr: Flow_Stack_Element_Ptr := Get_Last_Flow_Container;
      Last_Node_Ptr:       Nodes.Node_Class_Ptr   := null;
   begin
      if (Flow_Stack_Elem_Ptr /= null) then
         if (Flow_Stack_Elem_Ptr.Flow.all in CFGs.CFG_Object'Class) then
            if not Node_Stack.Empty(Flow_Stack_Elem_Ptr.NStack) then
               Node_Stack.Peek(Flow_Stack_Elem_Ptr.NStack, Last_Node_Ptr);
            end if;
         end if;
      end if;
      return Last_Node_Ptr;
   end Get_Last_Node;

   function Replace_Last_Node (Node_Ptr: in Nodes.Node_Class_Ptr) return Nodes.Node_Class_Ptr is
   begin
      return Replace_Last_Node_And_Relink(Node_Ptr, Get_Last_Node);
   end Replace_Last_Node;

   function Replace_Last_Node_And_Relink (Node_Ptr:      in Nodes.Node_Class_Ptr;
                                          Prev_Node_Ptr: in Nodes.Node_Class_Ptr)
                                         return Nodes.Node_Class_Ptr is
      Flow_Stack_Elem_Ptr: Flow_Stack_Element_Ptr    := Get_Last_Flow_Container;
      Last_Node_Ptr:       Nodes.Node_Class_Ptr      := null;
      Top_Flow_Ptr:        Flow_Types.Flow_Class_Ptr := null;
      Top_Cfg_Ptr:         CFGs.CFG_Class_Ptr        := null;
   begin
      if (Flow_Stack_Elem_Ptr /= null) then
         Top_Flow_Ptr := Flow_Stack_Elem_Ptr.Flow;
         if (Top_Flow_Ptr.all in CFGs.CFG_Object'Class) then
            Top_Cfg_Ptr := CFGs.CFG_Class_Ptr(Top_Flow_Ptr);
            if not Node_Stack.Empty(Flow_Stack_Elem_Ptr.NStack) then
               -- Remove the old Node from the Stack
               Node_Stack.Pop(Flow_Stack_Elem_Ptr.NStack, Last_Node_Ptr);
            end if;
            -- Push the new Node onto the Stack
            Node_Stack.Push(Flow_Stack_Elem_Ptr.NStack, Node_Ptr);
            -- Update the Links
            if Prev_Node_Ptr /= null then
               -- A previous Node has been specified: Delete its last succ-Link;
               -- This last succ-Link is to the Node we've just removed from the Stack;
               -- If the removed Node is same as the specified previous Node, it would make
               -- no Sense to remove a Link because a Node has no link on itself anyway;
               if (Prev_Node_Ptr /= Last_Node_Ptr) then
                  Nodes.Rem_Last_Succ(Prev_Node_Ptr.all);
                  Nodes.Free(Last_Node_Ptr);
               end if;
               -- Link to the new Node
               Nodes.Add_Succ(Prev_Node_Ptr.all, Flow_Types.Flow_Class_Ptr(Node_Ptr));
               Nodes.Add_Pred(Node_Ptr.all, Flow_Types.Flow_Class_Ptr(Prev_Node_Ptr));
            else
               if Last_Node_Ptr /= null then
                  -- A last Node exists *but* it was not specified as the current one's
                  -- ancestor; This can only mean that we must remove it;
                  Nodes.Free(Last_Node_Ptr);
               end if;
               -- No previous Nodes exist: Mark the new Node as CFG Head
               CFGs.Set_Head(Top_Cfg_Ptr.all, Node_Ptr);
            end if;
         end if;
      end if;
      return Last_Node_Ptr;
   end Replace_Last_Node_And_Relink;

   procedure Add_Idents_To_Last_Node (Idents: in out UStr_List.Object) is
      Node: Nodes.Node_Class_Ptr := Get_Last_Node;
   begin
      if Node /= null then
         -- Uncomment this Line to see what Identifiers get saved along the Way!
         -- Put_Line(Debug, "[Transformation] Saving Identifiers: " & UStr_List_Dump(Idents));
         Nodes.Add_Idents(Node.all, Idents);
      elsif UStr_List.Count(Idents) > 0 then
         -- Comment this Line to supress Identifier Warnings!
         Put_Line(Debug, "[Transformation] [DBG/WARN] " &
                  "No available Node to add following Idents to: " &
                  UStr_List_Dump(Idents));
      end if;
   end Add_Idents_To_Last_Node;

   -- * Param Stack related Methots

   procedure Open_Param_Root (Param: in Nodes.Node_Class_Ptr) is
      Prev_Param: Nodes.Node_Class_Ptr := null;
   begin
      Param_Stack.Peek(PStack, Prev_Param);
      Param_Stack.Push(PStack, Param);
      if Prev_Param /= null then
         -- This is a piggyback-Tree
         null;
      else
         -- This is a new Root, so save it
         Trans_Data.Last_PTree := Param;
      end if;
   end Open_Param_Root;

   procedure Open_Param_Non_Root (Param: in Nodes.Node_Class_Ptr) is
      Prev_Param: Nodes.Node_Class_Ptr := null;
   begin
      if (Param = null) then
         Put_Line(Error, "[Transformation] [ERR] Really _NASTY_ things happen to those who open NULL parameter nodes!");
      end if;
      Param_Stack.Peek(PStack, Prev_Param);
      Param_Stack.Push(PStack, Param);
      if Prev_Param /= null then
         -- Make the Links if possible
         Nodes.Add_Succ(Prev_Param.all, Flow_Types.Flow_Class_Ptr(Param));
         Nodes.Add_Pred(Param.all, Flow_Types.Flow_Class_Ptr(Prev_Param));
      else
         -- This is by some strange Reason the Root of a PTree
         Put_Line(Debug, "[Transformation] [DBG]" &
                  " This PTree Node shouldn't be Root unless Entry Guardian Expression!");
         Trans_Data.Last_PTree := Param;
      end if;
   end Open_Param_Non_Root;

   procedure Close_Last_Param is
      Curr_Node:              Nodes.Node_Class_Ptr        := null;
      Prev_Nodes, Next_Nodes: Flow_Class_Ptr_List.Object;
      Prev_Node,  Next_Node:  Flow_Types.Flow_Class_Ptr   := null;
      Var_Name:               Unbounded_String            := S2u("");
   begin
      if (not Param_Stack.Empty(PStack)) then
         Param_Stack.Pop(PStack, Curr_Node);

         -- Info: ** On-The-Fly Param Tree Compression
         --
         --       We are closing here a Node from the Parameter Tree; If this Node bears no
         --       Information (no explicit Information - through Variables and no implicit
         --       Information - through it's Position in the Successor List of its Predecessor)
         --       we can remove it completely;

         if ((Curr_Node.all in Nodes.Param_Node_Object) and then
             (String'(Nodes.Get_Name(Curr_Node.all)) = "")) then
            -- This is a Parameter Node with no Name
            if (Nodes.Get_Var_Name(Nodes.Param_Node_Object(Curr_Node.all)) = "") then
               -- No Variable either -> No explicit Information
               Prev_Nodes := Flow_Types.Get_Preds(Flow_Types.Flow_Object(Curr_Node.all));
               if (Flow_Class_Ptr_List.Count(Prev_Nodes) = 1) then
                  -- This Node is not the Parameter Tree Root
                  Prev_Node := Flow_Class_Ptr_List.Get_Head(Prev_Nodes);
                  Next_Nodes := Flow_Types.Get_Succs(Flow_Types.Flow_Object(Curr_Node.all));
                  if (Flow_Class_Ptr_List.Count(Next_Nodes) = 1) then
                     -- This Node has exactly one Child - it can be removed
                     Next_Node := Flow_Class_Ptr_List.Get_Head(Next_Nodes);
                     Flow_Types.Rem_Last_Succ(Prev_Node.all);
                     Flow_Types.Rem_Last_Pred(Next_Node.all);
                     Flow_Types.Add_Succ(Prev_Node.all, Next_Node);
                     Flow_Types.Add_Pred(Next_Node.all, Prev_Node);
                     Nodes.Free(Curr_Node);
                  end if; -- one Successor
               end if; -- not Root
            end if; -- no Variable
         end if; -- Parameter with no Name
      end if; -- Stack not empty
   end Close_Last_Param;

   procedure Close_All_Params is
   begin
      Param_Stack.Clear(PStack);
   end Close_All_Params;

   function Param_Tree_Empty return Boolean is
   begin
      return Param_Stack.Empty(PStack);
   end Param_Tree_Empty;

   function Get_Last_Param_Node return Nodes.Node_Class_Ptr is
      Param: Nodes.Node_Class_Ptr := null;
   begin
      if (not Param_Stack.Empty(PStack)) then
         Param_Stack.Peek(PStack, Param);
      end if;
      return Param;
   end Get_Last_Param_Node;

   function Get_Last_Param_Root return Nodes.Node_Class_Ptr is
   begin
      return Trans_Data.Last_PTree;
   end Get_Last_Param_Root;

   -- ** A U X I L I A R Y   F U N C T I O N S

   function First_Line (Element: Asis.Element) return String is
      Elem_Lines: Asis.Text.Line_List := Lines(Element);
   begin
      return W2s(Line_Image(Elem_Lines(Elem_Lines'First)));
   end First_Line;

   procedure Add_Idents_To_Type_Description (TDescr: in out Flow_Types.Type_Descr_Ptr;
                                             Idents: in out UStr_List.Object) is
      Full_Name: Unbounded_String := S2u("");
   begin
      if (TDescr /= null) then
         -- Walk the Name List tokenizing each found Identifier
         UStr_List.Reset(Idents);
         while UStr_List.Has_Next(Idents) loop
            UStr_List.Get_Next(Trans_Data.Names, Full_Name);
            UStr_List_List.Append(TDescr.Used_Idents, Tokenize(U2s(Full_Name), "."));
         end loop;
      elsif (not UStr_List.Empty(Idents)) then
         -- Comment this Line to supress Identifier Warnings!
         Put_Line(Debug, "[Transformation] [DBG/WARN] " &
                  "No available Type Structure to add following Idents to: " &
                  UStr_List_Dump(Idents));
      end if;
   end Add_Idents_To_Type_Description;

   procedure Save_Identifiers (Idents: in out UStr_List.Object) is
      Type_Ptr: Flow_Types.Type_Descr_Ptr := null;
   begin
      if UStr_List.Empty(Idents) then
         return;
      end if;
      -- Check for Type Description Pointers first!
      if (Trans_Data.Decl /= null) then
         Type_Ptr := Trans_Data.Decl.VType'Access;
      elsif (not Formal_Decl_Class_Ptr_List.Empty(Trans_Data.Form_Decls)) then
         Type_Ptr := Formal_Decl_Class_Ptr_List.Get_Tail(Trans_Data.Form_Decls).VType'Access;
      elsif (Trans_Data.Type_Decl /= null) then
         -- This must be the *LAST* Check!
         Type_Ptr := Trans_Data.Type_Decl;
      else
         Type_Ptr := null;
      end if;
      if Type_Ptr /= null then
         Add_Idents_To_Type_Description (Type_Ptr, Idents);
         return;
      end if;
      if Get_Last_Node /= null then
         Add_Idents_To_Last_Node (Idents);
         return;
      end if;
      Put_Line(Debug, "[Transformation] [DBG] Couldn't save Identifiers: " &
               UStr_List_Dump(Idents));
   end Save_Identifiers;

end Ast2Cfg.Transformation;
